#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib
import decimal
import copy

import octobot_commons.logging as logging

import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_trading.personal_data.positions.position_util as position_util
import octobot_trading.personal_data.positions.states as positions_states
import octobot_trading.personal_data.transactions.transaction_factory as transaction_factory
import octobot_trading.util as util
import octobot_trading.constants as constants


class Position(util.Initializable):
    def __init__(self, trader, symbol_contract):
        """
        When adding a new "dynamic" attribute, please add it to self.restore()
        """
        super().__init__()
        if self.is_inverse() is not symbol_contract.is_inverse_contract():
            raise errors.InvalidPosition(f"This position requires a "
                                         f"{'inverse' if symbol_contract.is_inverse_contract else 'linear'} contract")
        self.trader = trader
        self.exchange_manager = trader.exchange_manager
        self.simulated = trader.simulate

        self.logger_name = None
        self.position_id = None
        self.timestamp = 0
        self.symbol = None
        self.currency, self.market = None, None
        self.status = enums.PositionStatus.OPEN
        self.side = enums.PositionSide.UNKNOWN

        # Contract
        self.symbol_contract = symbol_contract

        # Prices
        self.entry_price = constants.ZERO
        self.exit_price = constants.ZERO
        self.mark_price = constants.ZERO
        self.liquidation_price = constants.ZERO
        self.fee_to_close = constants.ZERO

        # Size
        self.quantity = constants.ZERO
        self.size = constants.ZERO
        self.already_reduced_size = constants.ZERO
        self.value = constants.ZERO
        self.initial_margin = constants.ZERO
        self.margin = constants.ZERO

        # PNL
        self.unrealized_pnl = constants.ZERO
        self.realised_pnl = constants.ZERO

        # original position attributes
        self.creation_time = self.exchange_manager.exchange.get_exchange_current_time()

        # position state is initialized in initialize_impl()
        self.state = None

    @classmethod
    def get_name(cls):
        return cls.__name__

    def get_logger_name(self):
        """
        :return: The position logger name
        """
        if self.logger_name is None:
            self.logger_name = f"{self.get_name()} | {self.position_id}"
        return self.logger_name

    async def initialize_impl(self, **kwargs):
        """
        Initialize position status update tasks
        """
        await positions_states.create_position_state(self, **kwargs)

    async def on_open(self, force_open=False, is_from_exchange_data=False):
        """
        Triggers a new position open state
        :param force_open: if the new state should be forced
        :param is_from_exchange_data: if it's call from exchange data
        """
        self.state = positions_states.OpenPositionState(self, is_from_exchange_data=is_from_exchange_data)
        await self.state.initialize(forced=force_open)

    async def on_liquidate(self, force_liquidate=False, is_from_exchange_data=False):
        """
        Triggers a new position liquidation state
        :param force_liquidate: if the new state should be forced
        :param is_from_exchange_data: if it's call from exchange data
        """
        self.state = positions_states.LiquidatePositionState(self, is_from_exchange_data=is_from_exchange_data)
        await self.state.initialize(forced=force_liquidate)

    def _should_change(self, original_value, new_value):
        if new_value is not None and original_value != new_value:
            return True

    def _update(self, position_id, symbol, currency, market, timestamp,
                entry_price, mark_price, liquidation_price,
                quantity, size, value, initial_margin,
                unrealized_pnl, realised_pnl, fee_to_close,
                status=None):
        changed: bool = False

        if self._should_change(self.position_id, position_id):
            self.position_id = position_id

        if self._should_change(self.symbol, symbol):
            self.symbol, self.currency, self.market = symbol, currency, market

        if self._should_change(self.timestamp, timestamp):
            self.timestamp = timestamp
        if not self.timestamp:
            if not timestamp:
                self.creation_time = self.exchange_manager.exchange.get_exchange_current_time()
            else:
                # if we have a timestamp, it's a real trader => need to format timestamp if necessary
                self.creation_time = self.exchange_manager.exchange.get_uniformized_timestamp(timestamp)
            self.timestamp = self.creation_time

        if self._should_change(self.quantity, quantity):
            self.quantity = quantity
            changed = True

        if self._should_change(self.size, size):
            self.size = size
            changed = True

        if self._should_change(self.value, value):
            self.value = value
            changed = True

        if self._should_change(self.initial_margin, initial_margin):
            self.initial_margin = initial_margin
            self._update_margin()
            changed = True

        if self._should_change(self.unrealized_pnl, unrealized_pnl):
            self.unrealized_pnl = unrealized_pnl
            changed = True

        if self._should_change(self.realised_pnl, realised_pnl):
            self.realised_pnl = realised_pnl
            changed = True

        if self._should_change(self.entry_price, entry_price):
            self.entry_price = entry_price
            changed = True

        if self._should_change(self.mark_price, mark_price):
            self.mark_price = mark_price
            changed = True

        if self._should_change(self.fee_to_close, fee_to_close):
            self.fee_to_close = fee_to_close
            self._update_margin()
            changed = True

        if self._should_change(self.liquidation_price, liquidation_price):
            self.liquidation_price = liquidation_price
            changed = True

        if self._should_change(self.status.value, status):
            self.status = enums.PositionStatus(status)

        self._update_quantity_or_size_if_necessary()
        # update side after quantity as it relies on self.quantity
        self._update_side()
        self._update_prices_if_necessary(mark_price)
        return changed

    async def _ensure_position_initialized(self):
        """
        Checks if the position has already been initialized
        When it's not initialize it
        """
        if self.state is None:
            await self.initialize()

    async def update(self, update_size=None, mark_price=None, update_margin=None):
        """
        Updates position size and / or mark price
        :param update_size: the size update value
        :param mark_price: the mark price update value
        :param update_margin: the margin update value
        """
        await self._ensure_position_initialized()

        try:
            with self.update_or_restore():
                if mark_price is not None:
                    self._update_mark_price(mark_price)
                if update_margin is not None:
                    self._update_size_from_margin(update_margin)
                if update_size is not None:
                    self._update_size(update_size)
            if not self.is_idle():
                self._check_for_liquidation()
            else:
                await self.close()
        except errors.LiquidationPriceReached:
            self._update_exit_data(self.get_quantity_to_close(), self.mark_price)
            await self._create_liquidation_state()

    async def update_on_liquidation(self):
        """
        Update portfolio and position from a liquidation
        """
        size_update = self.get_quantity_to_close()
        self.unrealized_pnl = -self.initial_margin
        realised_pnl_update = self._update_realized_pnl_from_size_update(
            size_update, is_closing=True, update_price=self.mark_price,
            trigger_source=enums.PNLTransactionSource.LIQUIDATION)
        self._on_size_update(size_update, realised_pnl_update, self.unrealized_pnl, False)
        await self.close()

    def _update_mark_price(self, mark_price, check_liquidation=True):
        """
        Updates position mark_price and triggers size related attributes update
        :param mark_price: the update mark_price
        """
        self.mark_price = mark_price
        self._update_prices_if_necessary(mark_price)
        if check_liquidation:
            self._check_for_liquidation()
        if not self.is_idle() and self.exchange_manager.is_simulated:
            self.update_value()
            self.update_pnl()

    def _update_prices_if_necessary(self, mark_price):
        """
        Update the position entry price and mark price when their value is 0 or when the position is new
        :param mark_price: the current mark_price
        """
        if self.mark_price == constants.ZERO:
            self.mark_price = mark_price
        if self.entry_price == constants.ZERO:
            self.entry_price = mark_price

    def _update_size_from_margin(self, margin_update):
        """
        Updates position size and margin by converting margin to size and calling self._update_size
        :param margin_update: the position margin update
        """
        try:
            self._update_size(self.get_size_from_margin(margin_update))
        except (decimal.DivisionByZero, decimal.InvalidOperation):
            pass

    def is_order_increasing_size(self, order):
        """
        Check if an order increase the position size
        :param order: the order instance
        :return: True if the order increase the position size when filled
        """
        if order.reduce_only or order.close_position:
            return False  # Only reducing
        return (
                (self.is_idle() and self.symbol_contract.is_one_way_position_mode())
                or (self.is_long() and order.is_long())
                or (self.is_short() and order.is_short())
        )

    def update_from_order(self, order):
        """
        Update position size and entry price from filled order portfolio
        :param order: the filled order instance
        :return: the updated quantity, True if the order increased position size
        """
        # consider the order filled price as the current reference price for pnl computation
        # do not check liquidation as our position might be closed by this order
        self._update_mark_price(order.filled_price, check_liquidation=False)

        # get size to close to check if closing
        size_to_close = self.get_quantity_to_close()

        # Remove / add order fees from realized pnl
        realised_pnl_fees_update = self._update_realized_pnl_from_order(order)

        trigger_source = order_util.get_pnl_transaction_source_from_order(order)

        # Close position if order is closing position
        if order.close_position:
            # set position size to 0 to schedule position close at the next update
            self._update_size(size_to_close,
                              realised_pnl_update=realised_pnl_fees_update,
                              trigger_source=trigger_source)
            return

        # Calculates position quantity update from order
        size_update = self._calculates_size_update_from_filled_order(order, size_to_close)

        # set position entry price if necessary
        self._update_prices_if_necessary(order.filled_price)

        # Updates position average entry price from order only when increasing position side
        if self._is_update_increasing_size(size_update):
            if self.size == constants.ZERO:
                self.creation_time = self.exchange_manager.exchange.get_exchange_current_time()
            self.update_average_entry_price(size_update, order.filled_price)
        elif self._is_update_decreasing_size(size_update):
            self._update_exit_data(size_update, self.mark_price)

        # update size and realised pnl
        self._update_size(size_update, realised_pnl_update=realised_pnl_fees_update, trigger_source=trigger_source)

    def _update_realized_pnl_from_order(self, order):
        """
        Updates the position realized pnl from order
        Removes order's fees from realized pnl
        :param order: the realized pnl update
        """
        fees_currency = order.currency if self.symbol_contract.is_inverse_contract() else order.market
        realised_pnl_update = -order.get_total_fees(fees_currency)
        transaction_factory.create_fee_transaction(self.exchange_manager, fees_currency, self.symbol,
                                                   quantity=realised_pnl_update,
                                                   order_id=order.order_id)
        self.realised_pnl += realised_pnl_update
        return realised_pnl_update

    def _calculates_size_update_from_filled_order(self, order, size_to_close):
        """
        Calculates position size update from an order filled quantity
        :param order: the filled order
        :param size_to_close: the size to close the position
        :return: the position size update
        """
        order_quantity = order.filled_quantity if order.is_long() else -order.filled_quantity
        if order.reduce_only or not self.symbol_contract.is_one_way_position_mode():
            if self.is_long() and order.is_short():
                return max(order_quantity, size_to_close)
            if self.is_short() and order.is_long():
                return min(order_quantity, size_to_close)
            if not order.reduce_only:
                return order_quantity
            # Can't reduce position
            return constants.ZERO
        return order_quantity

    def _is_update_increasing_size(self, size_update):
        """
        :param size_update: the size update
        :return: True if this update will increase position size
        """
        if self.is_idle():
            return True
        if self.is_long():
            return size_update > constants.ZERO
        return size_update < constants.ZERO

    def _is_update_decreasing_size(self, size_update):
        """
        :param size_update: the size update
        :return: True if this update will increase position size
        """
        if self.is_idle():
            return False
        if self.is_long():
            return size_update < constants.ZERO
        return size_update > constants.ZERO

    def _is_update_closing(self, size_update):
        """
        :param size_update: the size update
        :return: True if this update will close the position
        """
        if self.is_idle():
            return True
        if self.is_long():
            return self.size + size_update <= constants.ZERO
        return self.size + size_update >= constants.ZERO

    def _update_size(self, size_update, realised_pnl_update=constants.ZERO,
                     trigger_source=enums.PNLTransactionSource.UNKNOWN):
        """
        Updates position size and triggers size related attributes update
        :param size_update: the size quantity
        :param realised_pnl_update: the current realised pnl update
        :return: True if the update increased position size
        """
        margin_update = constants.ZERO
        is_update_increasing_position_size = self._is_update_increasing_size(size_update)
        if self._is_update_decreasing_size(size_update):
            realised_pnl_update += self._update_realized_pnl_from_size_update(
                size_update, is_closing=self._is_update_closing(size_update),
                update_price=self.mark_price, trigger_source=trigger_source)
        self._check_and_update_size(size_update)
        self._update_quantity()
        self._update_side()
        if self.exchange_manager.is_simulated:
            margin_update = self._update_initial_margin()
            self.update_fee_to_close()
            self.update_liquidation_price()
            self.update_value()
            self.update_pnl()
        self._on_size_update(size_update,
                             realised_pnl_update,
                             margin_update,
                             is_update_increasing_position_size)

    def _update_realized_pnl_from_size_update(self, size_update, is_closing=False, update_price=constants.ZERO,
                                              trigger_source=enums.PNLTransactionSource.UNKNOWN):
        """
        Updates the position realized pnl from update size
        :param size_update: the position update size
        :param is_closing: True when the position will be closed after size update
        """
        try:
            realised_pnl_update = -size_update / self.size * self.unrealized_pnl
            transaction_factory.create_realised_pnl_transaction(self.exchange_manager,
                                                                self.get_currency(),
                                                                self.symbol,
                                                                self.side,
                                                                realised_pnl=realised_pnl_update,
                                                                is_closed_pnl=is_closing,
                                                                cumulated_closed_quantity=self.already_reduced_size,
                                                                closed_quantity=size_update,
                                                                first_entry_time=self.creation_time,
                                                                average_entry_price=self.entry_price,
                                                                average_exit_price=self.exit_price,
                                                                order_exit_price=update_price,
                                                                leverage=self.symbol_contract.current_leverage,
                                                                trigger_source=trigger_source)
        except (decimal.DivisionByZero, decimal.InvalidOperation):
            realised_pnl_update = constants.ZERO
        self.realised_pnl += realised_pnl_update
        return realised_pnl_update

    def _check_and_update_size(self, size_update):
        """
        Updates the position size with a valid size according to the contract position mode if not close the position
        This check is not mandatory when using one way position mode because it can't produce invalid size with
        :param size_update: the size update
        """
        if not self.symbol_contract.is_one_way_position_mode() and \
                ((self.is_long() and self.size + size_update < constants.ZERO) or
                 (self.is_short() and self.size + size_update > constants.ZERO)):
            self.size = constants.ZERO
        else:
            self.size += size_update

    def _update_quantity_or_size_if_necessary(self):
        """
        Set quantity from size if quantity is not set and size is set or update size
        """
        if self.quantity == constants.ZERO and self.size != constants.ZERO:
            self._update_quantity()
        elif self.size == constants.ZERO and self.quantity != constants.ZERO:
            self.size = self.quantity * self.symbol_contract.current_leverage

    def _update_quantity(self):
        """
        Update position quantity from position quantity
        """
        self.quantity = self.size / self.symbol_contract.current_leverage

    def update_value(self):
        raise NotImplementedError("update_value not implemented")

    def update_pnl(self):
        """
        Update position unrealised pnl
        """
        try:
            self.unrealized_pnl = self.get_unrealized_pnl(self.mark_price)
            self.on_pnl_update()
        except (decimal.DivisionByZero, decimal.InvalidOperation):
            self.unrealized_pnl = constants.ZERO

    def _update_exit_data(self, size_update, price):
        """
        Update position average exit price and already_reduced_size
        """
        try:
            self.update_average_exit_price(size_update, price)
        except (decimal.DivisionByZero, decimal.InvalidOperation):
            self.exit_price = constants.ZERO
        self.already_reduced_size += size_update

    def get_unrealized_pnl(self, price):
        raise NotImplementedError("get_unrealized_pnl not implemented")

    def get_margin_from_size(self, size):
        raise NotImplementedError("get_margin_from_size not implemented")

    def get_size_from_margin(self, margin):
        raise NotImplementedError("get_size_from_margin not implemented")

    def _update_initial_margin(self):
        """
        Updates position initial margin
        """
        margin_update = constants.ZERO
        try:
            previous_initial_margin = self.initial_margin
            self.initial_margin = self.get_margin_from_size(self.size).copy_abs()
            self._update_margin()
            margin_update = self.initial_margin - previous_initial_margin
        except (decimal.DivisionByZero, decimal.InvalidOperation):
            self.initial_margin = constants.ZERO
        return margin_update

    def update_average_entry_price(self, update_size, update_price):
        raise NotImplementedError("get_average_entry_price not implemented")

    def update_average_exit_price(self, update_size, update_price):
        raise NotImplementedError("update_average_exit_price not implemented")

    def get_initial_margin_rate(self):
        """
        :return: Initial Margin Rate = 1 / Leverage
        """
        try:
            return constants.ONE / self.symbol_contract.current_leverage
        except (decimal.DivisionByZero, decimal.InvalidOperation):
            return constants.ZERO

    def get_currency(self):
        """
        :return: position currency
        """
        return self.currency if self.symbol_contract.is_inverse_contract() else self.market

    def calculate_maintenance_margin(self):
        raise NotImplementedError("calculate_maintenance_margin not implemented")

    def update_liquidation_price(self):
        """
        Updates position liquidation price
        Should call _update_fee_to_close() at the end of the implementation
        """
        if self.symbol_contract.is_isolated():
            self.update_isolated_liquidation_price()
        else:
            self.update_cross_liquidation_price()

    def update_cross_liquidation_price(self):
        raise NotImplementedError("update_cross_liquidation_price not implemented")

    def update_isolated_liquidation_price(self):
        raise NotImplementedError("update_isolated_liquidation_price not implemented")

    def get_bankruptcy_price(self, price, side, with_mark_price=False):
        """
        The bankruptcy price refers to the price at which the initial margin of all positions is lost.
        :param price: the price to compute bankruptcy from
        :param side: the side of the position
        :param with_mark_price: if price should be mark price instead of entry price
        :return: the bankruptcy price
        """
        raise NotImplementedError("get_bankruptcy_price not implemented")

    def get_maker_fee(self, symbol):
        """
        :return: Position maker fee
        """
        try:
            symbol_fees = self.exchange_manager.exchange.get_fees(symbol)
            return decimal.Decimal(
                f"{symbol_fees[enums.ExchangeConstantsMarketPropertyColumns.MAKER.value]}")
        except (decimal.DivisionByZero, decimal.InvalidOperation):
            return constants.ZERO

    def get_taker_fee(self, symbol):
        """
        :return: Position taker fee
        """
        try:
            symbol_fees = self.exchange_manager.exchange.get_fees(symbol)
            return decimal.Decimal(
                f"{symbol_fees[enums.ExchangeConstantsMarketPropertyColumns.TAKER.value]}")
        except (decimal.DivisionByZero, decimal.InvalidOperation):
            return constants.ZERO

    def get_two_way_taker_fee_for_quantity_and_price(self, quantity, price, side, symbol):
        """
        # Fee to open = (Quantity of contracts * Order Price) x Taker fee
        # Fee to close = (Quantity of contracts * Bankruptcy Price derived from Order Price) x Taker fee
        :return: 2-way taker fee = fee to open + fee to close
        """
        return self.get_fee_to_open(quantity, price, symbol) + self.get_fee_to_close(quantity, price, side, symbol)

    def get_two_way_taker_fee(self):
        """
        :return: 2-way taker fee = fee to open + fee to close
        """
        return self.get_fee_to_open(self.size, self.mark_price, self.symbol) + self.fee_to_close

    def get_order_cost(self):
        raise NotImplementedError("get_order_cost not implemented")

    def get_fee_to_open(self, quantity, price, symbol):
        raise NotImplementedError("get_fee_to_open not implemented")

    def get_fee_to_close(self, quantity, price, side, symbol, with_mark_price=False):
        raise NotImplementedError("get_fee_to_close not implemented")

    def update_fee_to_close(self):
        raise NotImplementedError("update_fee_to_close not implemented")

    @staticmethod
    def is_inverse():
        raise NotImplementedError("is_inverse not implemented")

    def _update_margin(self):
        """
        Updates position margin = Initial margin + Fee to close
        """
        self.margin = self.initial_margin + self.fee_to_close

    def is_open(self):
        return self.state is None or self.state.is_open()

    def is_liquidated(self):
        return self.state is not None and self.state.is_liquidated()

    def is_refreshing(self):
        return self.state is not None and self.state.is_refreshing()

    def is_long(self):
        return self.side is enums.PositionSide.LONG

    def is_short(self):
        return self.side is enums.PositionSide.SHORT

    def is_idle(self):
        return self.quantity == constants.ZERO

    def get_quantity_to_close(self):
        """
        :return: the order quantity to close the position
        """
        return -self.size

    def get_unrealized_pnl_percent(self):
        """
        :return: Unrealized P&L% = [ Position's unrealized P&L / Position Margin ] x 100%
        """
        try:
            return (self.unrealized_pnl / self.margin) * constants.ONE_HUNDRED
        except (decimal.DivisionByZero, decimal.InvalidOperation):
            return constants.ZERO

    def on_pnl_update(self):
        """
        Triggers external calls when position pnl has been updated
        """
        self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.update_portfolio_from_pnl(self)

    def _on_side_update(self):
        """
        Resets the side related data when a position side changes
        """
        self._reset_entry_price()
        self.exit_price = constants.ZERO
        self.creation_time = self.exchange_manager.exchange.get_exchange_current_time()
        logging.get_logger(self.get_logger_name()).info(f"Changed position side: now {self.side.name}")

    def _on_size_update(self,
                        size_update,
                        realised_pnl_update,
                        margin_update,
                        is_update_increasing_position_size):
        """
        Triggers external calls when position size has been updated
        """
        self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio. \
            update_portfolio_data_from_position_size_update(self,
                                                            realised_pnl_update,
                                                            size_update,
                                                            margin_update,
                                                            is_update_increasing_position_size)

    async def recreate(self):
        """
        Recreate itself using its PositionManager instance
        """
        self.exchange_manager.exchange_personal_data.positions_manager.recreate_position(self)

    def update_from_raw(self, raw_position):
        symbol = str(raw_position.get(enums.ExchangeConstantsPositionColumns.SYMBOL.value, None))
        currency, market = self.exchange_manager.get_exchange_quote_and_base(symbol)
        # side is managed locally, do not parse it
        return self._update(
            symbol=symbol,
            currency=currency,
            market=market,
            entry_price=raw_position.get(enums.ExchangeConstantsPositionColumns.ENTRY_PRICE.value, constants.ZERO),
            mark_price=raw_position.get(enums.ExchangeConstantsPositionColumns.MARK_PRICE.value, constants.ZERO),
            liquidation_price=raw_position.get(enums.ExchangeConstantsPositionColumns.LIQUIDATION_PRICE.value,
                                               constants.ZERO),
            quantity=raw_position.get(enums.ExchangeConstantsPositionColumns.QUANTITY.value, constants.ZERO),
            size=raw_position.get(enums.ExchangeConstantsPositionColumns.SIZE.value, constants.ZERO),
            value=raw_position.get(enums.ExchangeConstantsPositionColumns.NOTIONAL.value, constants.ZERO),
            initial_margin=raw_position.get(enums.ExchangeConstantsPositionColumns.INITIAL_MARGIN.value,
                                            constants.ZERO),
            position_id=str(raw_position.get(enums.ExchangeConstantsPositionColumns.ID.value, symbol)),
            timestamp=raw_position.get(enums.ExchangeConstantsPositionColumns.TIMESTAMP.value, 0),
            unrealized_pnl=raw_position.get(enums.ExchangeConstantsPositionColumns.UNREALIZED_PNL.value,
                                            constants.ZERO),
            realised_pnl=raw_position.get(enums.ExchangeConstantsPositionColumns.REALISED_PNL.value, constants.ZERO),
            fee_to_close=raw_position.get(enums.ExchangeConstantsPositionColumns.CLOSING_FEE.value, constants.ZERO),
            status=position_util.parse_position_status(raw_position)
        )

    def to_dict(self):
        return {
            enums.ExchangeConstantsPositionColumns.ID.value: self.position_id,
            enums.ExchangeConstantsPositionColumns.SYMBOL.value: self.symbol,
            enums.ExchangeConstantsPositionColumns.STATUS.value: self.status.value,
            enums.ExchangeConstantsPositionColumns.TIMESTAMP.value: self.timestamp,
            enums.ExchangeConstantsPositionColumns.SIDE.value: self.side.value,
            enums.ExchangeConstantsPositionColumns.QUANTITY.value: self.quantity,
            enums.ExchangeConstantsPositionColumns.SIZE.value: self.size,
            enums.ExchangeConstantsPositionColumns.NOTIONAL.value: self.value,
            enums.ExchangeConstantsPositionColumns.INITIAL_MARGIN.value: self.initial_margin,
            enums.ExchangeConstantsPositionColumns.COLLATERAL.value: self.margin,
            enums.ExchangeConstantsPositionColumns.ENTRY_PRICE.value: self.entry_price,
            enums.ExchangeConstantsPositionColumns.MARK_PRICE.value: self.mark_price,
            enums.ExchangeConstantsPositionColumns.LIQUIDATION_PRICE.value: self.liquidation_price,
            enums.ExchangeConstantsPositionColumns.UNREALIZED_PNL.value: self.unrealized_pnl,
            enums.ExchangeConstantsPositionColumns.REALISED_PNL.value: self.realised_pnl,
        }

    def _check_for_liquidation(self):
        """
        _check_for_liquidation defines rules for a position to be liquidated
        """
        if (self.is_short()
            and self.mark_price >= self.liquidation_price > constants.ZERO) or (
            self.is_long()
            and self.mark_price <= self.liquidation_price > constants.ZERO
        ):
            raise errors.LiquidationPriceReached

    async def _create_liquidation_state(self):
        """
        create liquidation state
        """
        self.status = enums.PositionStatus.LIQUIDATING
        await positions_states.create_position_state(self)

    def _reset_entry_price(self):
        """
        Reset the entry price to ZERO and force entry price update
        """
        self.entry_price = constants.ZERO
        self._update_prices_if_necessary(self.mark_price)

    def _update_side(self):
        """
        Checks if self.side still represents the position side
        Only relevant when account is using one way position mode
        """
        if self.symbol_contract.is_one_way_position_mode() or self.side is enums.PositionSide.UNKNOWN:
            changed_side = False
            if self.quantity > constants.ZERO:
                if self.side is not enums.PositionSide.LONG:
                    self.side = enums.PositionSide.LONG
                    changed_side = True
            elif self.quantity < constants.ZERO:
                if self.side is not enums.PositionSide.SHORT:
                    self.side = enums.PositionSide.SHORT
                    changed_side = True
            else:
                self.side = enums.PositionSide.UNKNOWN
            if changed_side:
                self._on_side_update()

    def __str__(self):
        return self.to_string()

    def to_string(self):
        currency = self.get_currency()
        position_mode = self.symbol_contract.position_mode.value \
            if self.symbol_contract.position_mode else 'no position mode'
        return (f"{self.symbol} | "
                f"Size : {round(self.size, 10).normalize()} "
                f"({round(self.value, 10).normalize()} {currency}) "
                f"--> {self.side.value} {self.symbol_contract} | "
                f"Mark price : {round(self.mark_price, 10).normalize()} | "
                f"Entry price : {round(self.entry_price, 10).normalize()} | "
                f"Margin : {round(self.margin, 10).normalize()} {currency} | "
                f"Unrealized PNL : {round(self.unrealized_pnl, 14).normalize()} {currency} "
                f"({round(self.get_unrealized_pnl_percent(), 3)}%) | "
                f"Liquidation price : {round(self.liquidation_price, 10).normalize()} | "
                f"Realised PNL : {round(self.realised_pnl, 14).normalize()} {currency} "
                f"State : {self.state.state.value if self.state is not None else 'Unknown'} "
                f"({position_mode})")

    async def close(self):
        await self.reset()

    async def reset(self):
        """
        Reset position attributes
        """
        self.entry_price = constants.ZERO
        self.exit_price = constants.ZERO
        self.mark_price = constants.ZERO
        self.quantity = constants.ZERO
        self.size = constants.ZERO
        self.already_reduced_size = constants.ZERO
        self.value = constants.ZERO
        self.initial_margin = constants.ZERO
        self.margin = constants.ZERO
        self.liquidation_price = constants.ZERO
        self.fee_to_close = constants.ZERO
        self.status = enums.PositionStatus.OPEN
        self.side = enums.PositionSide.UNKNOWN
        self.unrealized_pnl = constants.ZERO
        self.realised_pnl = constants.ZERO
        self.creation_time = 0
        self.on_pnl_update()  # notify portfolio to reset unrealized PNL
        if not self.is_open():
            await self.on_open()

    def clear(self):
        """
        Clear position references
        """
        if self.state:
            self.state.clear()
        self.trader = None
        self.exchange_manager = None

    def restore(self, other_position):
        """
        Restore a position from another one
        :param other_position: the other position instance
        """
        self.entry_price = other_position.entry_price
        self.exit_price = other_position.exit_price
        self.mark_price = other_position.mark_price
        self.liquidation_price = other_position.liquidation_price
        self.fee_to_close = other_position.fee_to_close
        self.quantity = other_position.quantity
        self.size = other_position.size
        self.already_reduced_size = other_position.already_reduced_size
        self.value = other_position.value
        self.initial_margin = other_position.initial_margin
        self.margin = other_position.margin
        self.unrealized_pnl = other_position.unrealized_pnl
        self.realised_pnl = other_position.realised_pnl

    @contextlib.contextmanager
    def update_or_restore(self):
        """
        Ensure update complete without raising PortfolioNegativeValueError else restore Position instance's attributes
        """
        previous_position = copy.copy(self)
        try:
            yield
        except errors.PortfolioNegativeValueError:
            logging.get_logger(self.get_logger_name()).warning("Restoring after PortfolioNegativeValueError...")
            self.restore(previous_position)
            raise



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import pytest

import octobot_commons.constants as commons_constants
from octobot_backtesting.backtesting import Backtesting
from octobot_commons.asyncio_tools import wait_asyncio_next_cycle
from octobot_commons.tests.test_config import load_test_config
from octobot_trading.modes.channel.abstract_mode_consumer import AbstractTradingModeConsumer
from octobot_trading.enums import EvaluatorStates
import octobot_trading.constants as constants
import octobot_trading.errors as errors
from octobot_trading.exchanges.exchange_manager import ExchangeManager
from octobot_trading.modes import AbstractTradingMode
import octobot_trading.personal_data.portfolios.assets as portfolio_assets
from octobot_trading.exchanges.traders.trader_simulator import TraderSimulator
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def _get_tools():
    symbol = "BTC/USDT"
    config = load_test_config()
    config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_STARTING_PORTFOLIO]["SUB"] = \
        0.000000000000000000005
    config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_STARTING_PORTFOLIO]["BNB"] = \
        0.000000000000000000005
    config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_STARTING_PORTFOLIO]["USDT"] = 2000
    exchange_manager = ExchangeManager(config, "binance")

    # use backtesting not to spam exchanges apis
    exchange_manager.is_simulated = True
    exchange_manager.is_backtesting = True
    exchange_manager.backtesting = Backtesting(None, [exchange_manager.id], None, [])

    await exchange_manager.initialize()

    trader = TraderSimulator(config, exchange_manager)
    await trader.initialize()
    
    mode = AbstractTradingMode(config, exchange_manager)
    consumer = AbstractTradingModeConsumer(mode)

    return exchange_manager, symbol, consumer


async def test_can_create_order():
    _, symbol, consumer = await _get_tools()
    # portfolio: "BTC": 10 "USD": 1000
    not_owned_symbol = "ETH/BTC"
    not_owned_market = "BTC/ETH"
    min_trigger_symbol = "SUB/BTC"
    min_trigger_market = "ADA/BNB"

    # order from neutral state => false
    assert not await consumer.can_create_order(symbol, EvaluatorStates.NEUTRAL.value)

    # sell order using a currency with 0 available
    assert not await consumer.can_create_order(not_owned_symbol, EvaluatorStates.SHORT.value)
    assert not await consumer.can_create_order(not_owned_symbol, EvaluatorStates.VERY_SHORT.value)

    # sell order using a currency with < min available
    assert not await consumer.can_create_order(min_trigger_symbol, EvaluatorStates.SHORT.value)
    assert not await consumer.can_create_order(min_trigger_symbol, EvaluatorStates.VERY_SHORT.value)

    # sell order using a currency with > min available
    assert await consumer.can_create_order(not_owned_market, EvaluatorStates.SHORT.value)
    assert await consumer.can_create_order(not_owned_market, EvaluatorStates.VERY_SHORT.value)

    # buy order using a market with 0 available
    assert not await consumer.can_create_order(not_owned_market, EvaluatorStates.LONG.value)
    assert not await consumer.can_create_order(not_owned_market, EvaluatorStates.VERY_LONG.value)

    # buy order using a market with < min available
    assert not await consumer.can_create_order(min_trigger_market, EvaluatorStates.LONG.value)
    assert not await consumer.can_create_order(min_trigger_market, EvaluatorStates.VERY_LONG.value)

    # buy order using a market with > min available
    assert await consumer.can_create_order(not_owned_symbol, EvaluatorStates.LONG.value)
    assert await consumer.can_create_order(not_owned_symbol, EvaluatorStates.VERY_LONG.value)


async def test_can_create_order_unknown_symbols():
    _, _, consumer = await _get_tools()
    unknown_symbol = "VI?/BTC"
    unknown_market = "BTC/*s?"
    unknown_everything = "VI?/*s?"

    # buy order with unknown market
    assert not await consumer.can_create_order(unknown_market, EvaluatorStates.LONG.value)
    assert not await consumer.can_create_order(unknown_market, EvaluatorStates.VERY_LONG.value)
    assert await consumer.can_create_order(unknown_market, EvaluatorStates.SHORT.value)
    assert await consumer.can_create_order(unknown_market, EvaluatorStates.VERY_SHORT.value)

    # sell order with unknown symbol
    assert not await consumer.can_create_order(unknown_symbol, EvaluatorStates.SHORT.value)
    assert not await consumer.can_create_order(unknown_symbol, EvaluatorStates.VERY_SHORT.value)
    assert await consumer.can_create_order(unknown_symbol, EvaluatorStates.LONG.value)
    assert await consumer.can_create_order(unknown_symbol, EvaluatorStates.VERY_LONG.value)

    # neutral state with unknown symbol, market and everything
    assert not await consumer.can_create_order(unknown_symbol, EvaluatorStates.NEUTRAL.value)
    assert not await consumer.can_create_order(unknown_market, EvaluatorStates.NEUTRAL.value)
    assert not await consumer.can_create_order(unknown_everything,  EvaluatorStates.NEUTRAL.value)


async def test_valid_create_new_order():
    _, symbol, consumer = await _get_tools()

    # should raise NotImplementedError Exception
    with pytest.raises(NotImplementedError):
        await consumer.create_new_orders(symbol, -1, EvaluatorStates.NEUTRAL)
    with pytest.raises(NotImplementedError):
        await consumer.create_new_orders(symbol, -1, EvaluatorStates.VERY_SHORT, xyz=1)
    with pytest.raises(NotImplementedError):
        await consumer.create_new_orders(symbol, -1, EvaluatorStates.LONG, xyz=1, aaa="bbb")


async def test_get_holdings_ratio():
    exchange_manager, symbol, consumer = await _get_tools()
    exchange_manager.client_symbols = [symbol]
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
        last_prices_by_trading_pair[symbol] = decimal.Decimal("1000")
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
        portfolio_current_value = decimal.Decimal("11")
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio = {}
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio["BTC"] = \
        portfolio_assets.SpotAsset(name="BTC", available=decimal.Decimal("10"), total=decimal.Decimal("10"))
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio["USDT"] = \
        portfolio_assets.SpotAsset(name="USDT", available=decimal.Decimal("1000"), total=decimal.Decimal("1000"))

    assert consumer.get_holdings_ratio("BTC") == decimal.Decimal('0.9090909090909090909090909091')
    assert consumer.get_holdings_ratio("USDT") == decimal.Decimal('0.09090909090909090909090909091')

    exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio.pop("USDT")
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
        portfolio_current_value = decimal.Decimal("10")
    assert consumer.get_holdings_ratio("BTC") == constants.ONE
    # add ETH and try to get ratio without symbol price
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio.\
        get_currency_portfolio("ETH").total = decimal.Decimal(10)
    # force not backtesting mode
    exchange_manager.is_backtesting = False
    # force add symbol in exchange symbols
    exchange_manager.client_symbols.append("ETH/BTC")
    with pytest.raises(errors.MissingPriceDataError):
        ratio = consumer.get_holdings_ratio("ETH")
    # let channel register proceed
    await wait_asyncio_next_cycle()
    assert consumer.get_holdings_ratio("BTC") == constants.ONE
    assert consumer.get_holdings_ratio("USDT") == constants.ZERO
    assert consumer.get_holdings_ratio("XYZ") == constants.ZERO


async def test_get_number_of_traded_assets():
    exchange_manager, symbol, consumer = await _get_tools()
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
        origin_crypto_currencies_values = {
            symbol: 1,
            "xyz": 2,
            "aaa": 3
        }
    assert consumer.get_number_of_traded_assets() == 3



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.exchanges.abstract_websocket_exchange as abstract_websocket

cdef class AbstractWebsocketConnector(abstract_websocket.AbstractWebsocketExchange):
    cdef int timeout
    cdef int timeout_interval
    cdef int last_ping_time

    cdef bint is_connected
    cdef bint should_stop
    cdef bint is_authenticated
    cdef bint use_testnet

    cdef public dict endpoint_args

    # objects
    cdef public object websocket
    cdef object _watch_task
    cdef object last_msg

    cpdef void on_open(self)
    cpdef void on_auth(self, bint status)
    cpdef void on_close(self)
    cpdef void on_error(self, str error)
    cpdef void start(self)



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from tests import event_loop
from tests.exchanges import simulated_exchange_manager, simulated_trader

import octobot_trading.personal_data as personal_data

pytestmark = pytest.mark.asyncio


@pytest.fixture
def trade_manager_and_trader(simulated_trader):
    _, _, trader_instance = simulated_trader
    return personal_data.TradesManager(trader_instance), trader_instance


def test_has_closing_trade_with_order_id(trade_manager_and_trader):
    trade_manager, trader = trade_manager_and_trader
    assert trade_manager.has_closing_trade_with_order_id(None) is False
    assert trade_manager.has_closing_trade_with_order_id("None") is False
    trade = personal_data.Trade(trader)
    trade.trade_id = "id"
    trade.is_closing_order = False
    trade.origin_order_id = "None"
    trade_manager.trades["id"] = trade
    # trade is not closing order not has the right origin_order_id
    assert trade_manager.has_closing_trade_with_order_id("id") is False
    # trade does not has the right origin_order_id
    trade.is_closing_order = True
    assert trade_manager.has_closing_trade_with_order_id("id2") is False
    assert trade_manager.has_closing_trade_with_order_id("id") is False
    trade.origin_order_id = "id"
    # trade is closing this order
    assert trade_manager.has_closing_trade_with_order_id("id") is True



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import async_channel.constants as channel_constants

import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.enums as enums


class TradesProducer(exchanges_channel.ExchangeChannelProducer):
    async def push(self, trades, old_trade=False):
        await self.perform(trades, old_trade=old_trade)

    async def perform(self, trades, old_trade=False):
        try:
            for trade in trades:
                if not trade:
                    continue
                symbol: str = self.channel.exchange_manager.get_exchange_symbol(
                    trade[enums.ExchangeConstantsOrderColumns.SYMBOL.value])
                if self.channel.get_filtered_consumers(symbol=channel_constants.CHANNEL_WILDCARD) or \
                        self.channel.get_filtered_consumers(symbol=symbol):
                    trade_id: str = trade[enums.ExchangeConstantsOrderColumns.ID.value]

                    await self.channel.exchange_manager.exchange_personal_data.handle_trade_update(
                        symbol,
                        trade_id,
                        trade,
                        is_old_trade=old_trade,
                        should_notify=True)
        except asyncio.CancelledError:
            self.logger.info("Update tasks cancelled.")
        except Exception as e:
            self.logger.exception(e, True, f"Exception when triggering update: {e}")

    async def send(self, cryptocurrency, symbol, trade, old_trade=False):
        for consumer in self.channel.get_filtered_consumers(symbol=symbol):
            await consumer.queue.put({
                "exchange": self.channel.exchange_manager.exchange_name,
                "exchange_id": self.channel.exchange_manager.id,
                "cryptocurrency": cryptocurrency,
                "symbol": symbol,
                "trade": trade,
                "old_trade": old_trade
            })


class TradesChannel(exchanges_channel.ExchangeChannel):
    PRODUCER_CLASS = TradesProducer
    CONSUMER_CLASS = exchanges_channel.ExchangeChannelConsumer



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library
import octobot_commons.display as commons_display

import octobot_trading.exchange_channel as exchanges_channel


class AbstractStorage:
    IS_LIVE_CONSUMER = True
    USE_LIVE_CONSUMER_IN_BACKTESTING = False
    LIVE_CHANNEL = None
    IS_HISTORICAL = True
    HISTORY_TABLE = None

    def __init__(self, exchange_manager, plot_settings: commons_display.PlotSettings,
                 use_live_consumer_in_backtesting=None, is_historical=None):
        self.exchange_manager = exchange_manager
        self.plot_settings: commons_display.PlotSettings = plot_settings
        self.consumer = None
        self.use_live_consumer_in_backtesting = use_live_consumer_in_backtesting \
            or self.USE_LIVE_CONSUMER_IN_BACKTESTING
        self.is_historical = is_historical or self.IS_HISTORICAL
        self.enabled = True

    def should_register_live_consumer(self):
        return self.IS_LIVE_CONSUMER and \
           (
                not self.exchange_manager.is_backtesting or
                (self.exchange_manager.is_backtesting and self.use_live_consumer_in_backtesting)
            )

    async def start(self):
        if self.should_register_live_consumer():
            await self.register_live_consumer()
        await self.on_start()

    async def on_start(self):
        """
        Called after start, implement in necessary
        """

    async def register_live_consumer(self):
        if self.LIVE_CHANNEL is None:
            raise ValueError(f"self.live_channel has to be set")
        self.consumer = await exchanges_channel.get_chan(
            self.LIVE_CHANNEL,
            self.exchange_manager.id
        ).new_consumer(
            self._live_callback
        )

    async def enable(self, enabled):
        if self.enabled != enabled:
            self.enabled = enabled
            if self.enabled:
                await self.start()
            else:
                await self.stop(clear=False)

    async def stop(self, clear=True):
        if self.consumer is not None:
            await self.consumer.stop()
        if clear:
            self.consumer = None
            self.exchange_manager = None

    async def store_history(self):
        if self.enabled:
            await self._store_history()

    async def _live_callback(self, *args, **kwargs):
        raise NotImplementedError(f"_live_callback not implemented for {self.__class__.__name__}")

    async def _store_history(self):
        raise NotImplementedError(f"_store_history not implemented for {self.__class__.__name__}")

    def _get_db(self, *args):
        raise NotImplementedError(f"_get_db not implemented for {self.__class__.__name__}")

    async def clear_history(self, flush=True):
        if self.HISTORY_TABLE is None:
            raise NotImplementedError(f"{self.__class__.__name__}.HISTORY_TABLE has to be set")
        database = self._get_db()
        await database.delete(self.HISTORY_TABLE, None)
        if flush:
            await database.flush()



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.personal_data.orders.orders_manager as orders_manager
cimport octobot_trading.personal_data.orders.order as order_class
cimport octobot_trading.personal_data.portfolios.portfolio_manager as portfolio_manager
cimport octobot_trading.personal_data.positions.positions_manager as positions_manager
cimport octobot_trading.personal_data.trades.trades_manager as trades_manager
cimport octobot_trading.personal_data.transactions.transactions_manager as transactions_manager
cimport octobot_trading.util as util

cdef class ExchangePersonalData(util.Initializable):
    cdef public object logger
    cdef public object exchange

    cdef public dict config

    cdef public object exchange_manager
    cdef public object trader

    cdef public portfolio_manager.PortfolioManager portfolio_manager
    cdef public trades_manager.TradesManager trades_manager
    cdef public orders_manager.OrdersManager orders_manager
    cdef public positions_manager.PositionsManager positions_manager
    cdef public transactions_manager.TransactionsManager transactions_manager

    cpdef void clear(self)

    cdef bint _is_out_of_sync_order(self, str order_id)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy

import octobot_commons.symbols
import octobot_commons.time_frame_manager as time_frame_manager
import octobot_commons.constants as constants
import octobot_commons.logging as logging
import octobot_commons.errors as errors
import octobot_commons.enums as commons_enums
import octobot_commons.tree as commons_tree

import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.exchanges.config.backtesting_exchange_config as backtesting_exchange_config
import octobot_trading.constants as trading_constants
import octobot_trading.util as util


class ExchangeConfig(util.Initializable):
    def __init__(self, exchange_manager):
        super().__init__()
        self._logger = logging.get_logger(self.__class__.__name__)

        self.exchange_manager = exchange_manager
        self.config = exchange_manager.config

        # dict of exchange supported pairs by enabled currencies from self.config
        self.traded_cryptocurrencies = {}

        # list of exchange supported enabled pairs from self.config
        self.traded_symbol_pairs = []
        # Same as traded_symbol_pairs but with parsed symbols
        self.traded_symbols = []

        # list of exchange supported pairs from self.config (used to initialize evaluators and more)
        self.all_config_symbol_pairs = []

        # list of exchange supported pairs on which we want to collect data through updaters or websocket
        self.watched_pairs = []

        # list of required time frames from configuration that are available
        self.available_required_time_frames = []

        # list of exchange supported time frames that are also required (config + real time)
        self.traded_time_frames = []

        # list of time frames to be used for real-time purposes (short time frames)
        self.real_time_time_frames = []

        # number of required historical candles
        self.required_historical_candles_count = constants.DEFAULT_IGNORED_VALUE

        self.backtesting_exchange_config = None

    async def initialize_impl(self):
        pass

    def set_config_traded_pairs(self):  # TODO
        self._set_config_traded_pairs()

    def set_config_time_frame(self):  # TODO
        self._set_config_time_frame()

    def init_backtesting_exchange_config(self):
        self.backtesting_exchange_config = backtesting_exchange_config.BacktestingExchangeConfig()

    def set_historical_settings(self):
        self.required_historical_candles_count = self.config.get(constants.CONFIG_TENTACLES_REQUIRED_CANDLES_COUNT,
                                                                 constants.DEFAULT_IGNORED_VALUE)

    def get_shortest_time_frame(self):
        return self.traded_time_frames[-1]

    def initialize_exchange_event_tree(self):
        tree_provider = commons_tree.EventProvider.instance()
        for topic in trading_constants.DEFAULT_FUTURES_INITIALIZATION_EVENT_TOPICS \
                if self.exchange_manager.is_future \
                else trading_constants.DEFAULT_INITIALIZATION_EVENT_TOPICS:
            if topic in (
                    commons_enums.InitializationEventExchangeTopics.POSITIONS,
                    commons_enums.InitializationEventExchangeTopics.TRADES,
                    commons_enums.InitializationEventExchangeTopics.ORDERS,
                    commons_enums.InitializationEventExchangeTopics.CONTRACTS,
                    commons_enums.InitializationEventExchangeTopics.CANDLES,
                    commons_enums.InitializationEventExchangeTopics.PRICE,
                    commons_enums.InitializationEventExchangeTopics.ORDER_BOOK,
                    commons_enums.InitializationEventExchangeTopics.FUNDING,
            ):
                for symbol in self.traded_symbol_pairs:
                    if topic in (
                            commons_enums.InitializationEventExchangeTopics.CANDLES,
                    ):
                        for time_frame in self.traded_time_frames:
                            tree_provider.create_event_at_path(
                                self.exchange_manager.bot_id, commons_tree.get_exchange_path(
                                    self.exchange_manager.exchange_name,
                                    topic.value,
                                    symbol=symbol,
                                    time_frame=time_frame.value
                                )
                            )
                    else:
                        tree_provider.create_event_at_path(
                            self.exchange_manager.bot_id, commons_tree.get_exchange_path(
                                self.exchange_manager.exchange_name,
                                topic.value,
                                symbol=symbol
                            )
                        )
            else:
                tree_provider.create_event_at_path(
                    self.exchange_manager.bot_id, commons_tree.get_exchange_path(
                        self.exchange_manager.exchange_name,
                        topic.value
                    )
                )

    def get_relevant_time_frames(self):
        # If required timeframes: use those. Use traded timeframes otherwise
        return self.available_required_time_frames or self.traded_time_frames

    async def add_watched_symbols(self, symbols):
        new_valid_symbols = [
            symbol
            for symbol in symbols
            if self._is_valid_symbol(symbol, []) and symbol not in self.watched_pairs
        ]
        try:
            if new_valid_symbols:
                self.watched_pairs += new_valid_symbols
                self._logger.debug(f"Adding watched symbols: {new_valid_symbols}")
                if self.exchange_manager.has_websocket:
                    self.exchange_manager.exchange_web_socket.add_pairs(new_valid_symbols, watching_only=True)
                    await self.exchange_manager.exchange_web_socket.close_and_restart_sockets(debounce_duration=1)

                await exchange_channel.get_chan(trading_constants.TICKER_CHANNEL,
                                                self.exchange_manager.id).modify(added_pairs=new_valid_symbols)
        except Exception as e:
            self._logger.error(f"Failed to add new watched symbols {symbols} : {e}")

    def _set_config_traded_pairs(self):
        self.traded_cryptocurrencies = {}
        traded_symbol_pairs_set = set()
        existing_pairs = set()
        for cryptocurrency in self.config[constants.CONFIG_CRYPTO_CURRENCIES]:
            traded_symbol_pairs_set = self._set_config_traded_pair(cryptocurrency,
                                                                   traded_symbol_pairs_set,
                                                                   existing_pairs)
        self.traded_symbol_pairs = list(traded_symbol_pairs_set)
        self.traded_symbols = [
            octobot_commons.symbols.parse_symbol(symbol)
            for symbol in traded_symbol_pairs_set
        ]
        self.all_config_symbol_pairs = list(existing_pairs)

        # only add self.traded_symbol_pairs to watched pairs as not every existing_pairs are being collected
        self.watched_pairs = copy.deepcopy(self.traded_symbol_pairs)

    def _set_config_traded_pair(self, cryptocurrency, traded_symbol_pairs_set, existing_pairs):
        try:
            is_enabled = util.is_currency_enabled(self.config, cryptocurrency, True)
            if self.config[constants.CONFIG_CRYPTO_CURRENCIES][cryptocurrency][constants.CONFIG_CRYPTO_PAIRS]:
                if self.config[constants.CONFIG_CRYPTO_CURRENCIES][cryptocurrency][constants.CONFIG_CRYPTO_PAIRS] != \
                        constants.CONFIG_SYMBOLS_WILDCARD:
                    self._populate_non_wildcard_pairs(cryptocurrency, existing_pairs, is_enabled)
                else:
                    self._populate_wildcard_pairs(cryptocurrency, existing_pairs, is_enabled)
                # add to global traded pairs
                if is_enabled:
                    if not self.traded_cryptocurrencies[cryptocurrency]:
                        self._logger.error(
                            f"{self.exchange_manager.exchange_name} is not supporting any {cryptocurrency} trading pair"
                            f" from the current configuration.")
                    traded_symbol_pairs_set = traded_symbol_pairs_set.union(
                        self.traded_cryptocurrencies[cryptocurrency]
                    )
            elif is_enabled:
                self._logger.error(f"Current configuration for {cryptocurrency} is not including any trading pair, "
                                   f"this asset can't be traded and related orders won't be loaded. "
                                   f"OctoBot requires at least one trading pair in configuration to handle an asset. "
                                   f"You can add trading pair(s) for each asset in the configuration section.")
        except errors.ConfigError as err:
            self._logger.error(str(err))
        return traded_symbol_pairs_set

    def _populate_non_wildcard_pairs(self, cryptocurrency, existing_pairs, is_enabled):
        if self.config[constants.CONFIG_CRYPTO_CURRENCIES][cryptocurrency][constants.CONFIG_CRYPTO_PAIRS] != \
                constants.CONFIG_SYMBOLS_WILDCARD:
            currency_pairs = []
            for symbol in self.config[constants.CONFIG_CRYPTO_CURRENCIES][cryptocurrency][
                constants.CONFIG_CRYPTO_PAIRS]:
                if self.exchange_manager.symbol_exists(symbol):
                    if is_enabled:
                        currency_pairs.append(symbol)
                    # also add disabled pairs to existing pairs since they still exist on exchange
                    existing_pairs.add(symbol)
                elif is_enabled:
                    additional_details = ""
                    if self.exchange_manager.is_sandboxed:
                        additional_details = f" Exchange sandbox is enabled, please make sure this pair is traded on " \
                                             f" the {self.exchange_manager.exchange_name} sandbox as sandboxes " \
                                             f"usually only support a subset of the real exchange's pairs."
                    self._logger.error(f"{self.exchange_manager.exchange_name} is not supporting the "
                                       f"{symbol} trading pair.{additional_details}")
            if is_enabled:
                self.traded_cryptocurrencies[cryptocurrency] = currency_pairs

    def _populate_wildcard_pairs(self, cryptocurrency, existing_pairs, is_enabled):
        try:
            wildcard_pairs_list = self._create_wildcard_symbol_list(self.config[constants.CONFIG_CRYPTO_CURRENCIES]
                                                                    [cryptocurrency][constants.CONFIG_CRYPTO_QUOTE])
        except KeyError as e:
            raise errors.ConfigError(f"Impossible to use a wildcard configuration for {cryptocurrency}: missing {e} "
                                     f"value in {cryptocurrency} {constants.CONFIG_CRYPTO_CURRENCIES} "
                                     f"configuration.")

        # additional pairs
        if constants.CONFIG_CRYPTO_ADD in self.config[constants.CONFIG_CRYPTO_CURRENCIES][cryptocurrency]:
            wildcard_pairs_list += self._add_tradable_symbols_from_config(cryptocurrency,
                                                                          wildcard_pairs_list)

        if is_enabled:
            self.traded_cryptocurrencies[cryptocurrency] = wildcard_pairs_list

        # also add disabled pairs to existing pairs since they still exist on exchange
        existing_pairs.update(wildcard_pairs_list)

    def _set_config_time_frame(self):
        for time_frame in time_frame_manager.get_config_time_frame(self.config):
            if self.exchange_manager.time_frame_exists(time_frame.value):
                self.available_required_time_frames.append(time_frame)
        if not self.exchange_manager.is_backtesting or not self.available_required_time_frames:
            # add shortest time frame for realtime evaluators or if no time frame at all has
            # been registered in backtesting
            client_shortest_time_frame = time_frame_manager.find_min_time_frame(
                self.exchange_manager.client_time_frames,
                constants.MIN_EVAL_TIME_FRAME)
            self.real_time_time_frames.append(client_shortest_time_frame)
        self.available_required_time_frames = time_frame_manager.sort_time_frames(self.available_required_time_frames,
                                                                                  reverse=True)
        self.traded_time_frames = list(set().union(self.available_required_time_frames, self.real_time_time_frames))
        self.traded_time_frames = time_frame_manager.sort_time_frames(self.traded_time_frames, reverse=True)

    @staticmethod
    def _is_tradable_with_cryptocurrency(symbol, cryptocurrency):
        return symbol if octobot_commons.symbols.parse_symbol(symbol).quote == cryptocurrency else None

    def _add_tradable_symbols_from_config(self, cryptocurrency, filtered_symbols):
        return [
            symbol
            for symbol in self.config[constants.CONFIG_CRYPTO_CURRENCIES][cryptocurrency][constants.CONFIG_CRYPTO_ADD]
            if self._is_valid_symbol(symbol, filtered_symbols)
        ]

    def _is_valid_symbol(self, symbol, filtered_symbols):
        return self.exchange_manager.symbol_exists(symbol) and symbol not in filtered_symbols

    def _create_wildcard_symbol_list(self, cryptocurrency):
        return [s for s in [ExchangeConfig._is_tradable_with_cryptocurrency(symbol, cryptocurrency)
                            for symbol in self.exchange_manager.client_symbols]
                if s is not None]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.portfolios.assets cimport future_asset
from octobot_trading.personal_data.portfolios.assets cimport margin_asset
from octobot_trading.personal_data.portfolios.assets cimport spot_asset

from octobot_trading.personal_data.portfolios.assets.future_asset cimport (
    FutureAsset,
)
from octobot_trading.personal_data.portfolios.assets.margin_asset cimport (
    MarginAsset,
)
from octobot_trading.personal_data.portfolios.assets.spot_asset cimport (
    SpotAsset,
)

__all__ = [
    "FutureAsset",
    "MarginAsset",
    "SpotAsset",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import decimal
import logging
import os
import threading

import cryptofeed
import cryptofeed.callback as cryptofeed_callbacks
import cryptofeed.config as cryptofeed_config
import cryptofeed.connection_handler
import cryptofeed.defines as cryptofeed_constants
import cryptofeed.exchanges as cryptofeed_exchanges
import cryptofeed.types as cryptofeed_types

import octobot_commons
import octobot_commons.asyncio_tools as asyncio_tools
import octobot_commons.enums as commons_enums
import octobot_commons.logging as commons_logging
import octobot_commons.symbols as symbol_util
import octobot_commons.time_frame_manager as time_frame_manager

import octobot_trading.constants as trading_constants
import octobot_trading.enums as trading_enums
import octobot_trading.exchanges.abstract_websocket_exchange as abstract_websocket
import octobot_trading.exchanges.connectors.abstract_websocket_connector as abstract_websocket_connector
from octobot_trading.enums import ExchangeConstantsOrderBookInfoColumns as ECOBIC, \
    ExchangeConstantsTickersColumns as Ectc
from octobot_trading.enums import WebsocketFeeds as Feeds


class CryptofeedWebsocketConnector(abstract_websocket.AbstractWebsocketExchange):
    CRYPTOFEED_LOGGERS = ["feedhandler"] + abstract_websocket_connector.AbstractWebsocketConnector.LOGGERS
    CRYPTOFEED_DEFAULT_MARKET_SEPARATOR = "-"

    INIT_REQUIRING_EXCHANGE_FEEDS = [Feeds.CANDLE]

    IGNORED_FEED_PAIRS = {
        # When ticker or future index is available : no need to calculate mark price from recent trades
        Feeds.TRADES: [Feeds.TICKER, Feeds.FUTURES_INDEX],
        # When candles are available : use min timeframe kline to push ticker
        Feeds.TICKER: [Feeds.KLINE]
    }

    CRYPTOFEED_FEEDS_TO_WEBSOCKET_FEEDS = {
        # Unauthenticated
        cryptofeed_constants.TRADES: Feeds.TRADES,
        cryptofeed_constants.TICKER: Feeds.TICKER,
        cryptofeed_constants.CANDLES: Feeds.CANDLE,
        cryptofeed_constants.L1_BOOK: Feeds.L1_BOOK,
        cryptofeed_constants.L2_BOOK: Feeds.L2_BOOK,
        cryptofeed_constants.L3_BOOK: Feeds.L3_BOOK,
        cryptofeed_constants.FUNDING: Feeds.FUNDING,
        cryptofeed_constants.LIQUIDATIONS: Feeds.LIQUIDATIONS,
        cryptofeed_constants.OPEN_INTEREST: Feeds.OPEN_INTEREST,
        cryptofeed_constants.INDEX: Feeds.FUTURES_INDEX,

        # Authenticated
        cryptofeed_constants.TRANSACTIONS: Feeds.TRANSACTIONS,
        cryptofeed_constants.BALANCES: Feeds.PORTFOLIO,
        cryptofeed_constants.PLACE_ORDER: Feeds.ORDERS,
        cryptofeed_constants.CANCEL_ORDER: Feeds.ORDERS,
        cryptofeed_constants.ORDERS: Feeds.ORDERS,
        cryptofeed_constants.ORDER_STATUS: Feeds.ORDERS,
        cryptofeed_constants.ORDER_INFO: Feeds.ORDERS,
        cryptofeed_constants.TRADE_HISTORY: Feeds.TRADE,
        cryptofeed_constants.FILLS: Feeds.TRADE,
        cryptofeed_constants.POSITIONS: Feeds.POSITION,
    }

    PAIR_INDEPENDENT_CHANNELS = [
        cryptofeed_constants.BALANCES,
        cryptofeed_constants.ORDER_INFO,
    ]
    CANDLE_CHANNELS = [
        cryptofeed_constants.CANDLES,
    ]
    WATCHED_PAIR_CHANNELS = [
        cryptofeed_constants.TRADES,
        cryptofeed_constants.TICKER,
        cryptofeed_constants.CANDLES,
    ]

    EXCHANGE_CONSTRUCTOR_KWARGS = {}

    def __init__(self, config: object, exchange_manager: object):
        super().__init__(config, exchange_manager)
        self.channels = []

        self.callbacks = {}
        self.candle_callback = None

        self.filtered_pairs = []
        self.watched_pairs = []
        self.min_timeframe = None

        self.local_loop = None
        self.is_websocket_restarting = False

        self._fix_signal_handler()

        # Manage cryptofeed loggers
        self.client_logger = logging.getLogger(f"WebSocketClient - {self.name}")
        self.client_logger.setLevel(logging.WARNING)
        self._fix_logger()

        self.client = None
        self.client_config = None
        commons_logging.set_logging_level(self.CRYPTOFEED_LOGGERS, logging.WARNING)

        self.callback_by_feed = {
            # Unauthenticated
            cryptofeed_constants.TRADES: cryptofeed_callbacks.TradeCallback(self.trades),
            cryptofeed_constants.TICKER: cryptofeed_callbacks.TickerCallback(self.ticker),
            cryptofeed_constants.CANDLES: cryptofeed_callbacks.CandleCallback(self.candle),
            cryptofeed_constants.FUNDING: cryptofeed_callbacks.FundingCallback(self.funding),
            cryptofeed_constants.OPEN_INTEREST: cryptofeed_callbacks.OpenInterestCallback(self.open_interest),
            cryptofeed_constants.LIQUIDATIONS: cryptofeed_callbacks.LiquidationCallback(self.liquidations),
            cryptofeed_constants.INDEX: cryptofeed_callbacks.IndexCallback(self.index),
            cryptofeed_constants.L2_BOOK: cryptofeed_callbacks.BookCallback(self.book),

            # Authenticated
            cryptofeed_constants.ORDER_INFO: cryptofeed_callbacks.OrderInfoCallback(self.order),
            cryptofeed_constants.TRANSACTIONS: cryptofeed_callbacks.TransactionsCallback(self.transaction),
            cryptofeed_constants.BALANCES: cryptofeed_callbacks.BalancesCallback(self.balance),
            cryptofeed_constants.FILLS: cryptofeed_callbacks.UserFillsCallback(self.fill),
        }
        self._set_async_callbacks()

        # Create cryptofeed FeedHandler instance
        self._create_client()

        # Creates cryptofeed exchange instance
        self.cryptofeed_exchange = cryptofeed_exchanges.EXCHANGE_MAP[self.get_feed_name()](
            config=self.client_config,
            **self._get_feed_default_kwargs()
        )

        self._previous_open_candles = {}
        self._full_symbol_by_feed_symbol = {}

    """
    Abstract methods
    """

    @classmethod
    def get_feed_name(cls):
        raise NotImplementedError("get_feed_name not implemented")

    """
    Abstract implementations
    """

    def start(self):
        self._init_client()
        self._start_client()

    async def _inner_stop(self):
        try:
            for feed in self.client.feeds:
                feed.stop()
            for feed in self.client.feeds:
                await feed.shutdown()
            if self.client.raw_data_collection:
                self.client.raw_data_collection.stop()
        except Exception as e:
            self.logger.exception(e, False)
            self.logger.error(f"Failed to stop websocket feed : {e}")

    async def stop(self):
        """
        Reimplementation of self.client.stop() without calling loop.run_until_complete()
        """
        if asyncio.get_event_loop() is self.local_loop:
            await self._inner_stop()
        else:
            asyncio_tools.run_coroutine_in_asyncio_loop(self._inner_stop(), self.local_loop)

    async def close(self):
        """
        Can't call self.client.close() because it uses loop operations
        """
        try:
            self.client = None
        except Exception as e:
            self.logger.error(f"Failed to close websocket feed : {e}")

    def add_pairs(self, pairs, watching_only=False):
        """
        Add new pairs to self.filtered_pairs
        :param pairs: the list of pair to add
        :param watching_only: if pairs are for watching or trading purpose
        """
        for pair in pairs:
            self._add_pair(self.get_exchange_pair(pair), watching_only=watching_only)

    async def _inner_reset(self):
        """
        Removes and stops all running feeds an recreate them
        """
        try:
            self._remove_all_feeds()
            await self.stop()
            await self.close()

            self._create_client()
            if self.candle_callback is not None:
                self._subscribe_candle_feed()
            self._subscribe_channels_feeds()
            self._start_client(should_create_loop=False)
        except Exception as e:
            self.logger.error(f"Failed to reconnect to websocket : {e}")

    def _call_inner_reset(self):
        self.is_websocket_restarting = True
        try:
            asyncio_tools.run_coroutine_in_asyncio_loop(self._inner_reset(), self.local_loop)
        except Exception as e:
            self.logger.exception(e, True, f"Error when resetting websockets {e}")
        finally:
            self.is_websocket_restarting = False

    async def reset(self):
        if self.is_websocket_restarting:
            self.logger.debug("Reset attempt but a reset is already in progress.")
            return
        # force is_websocket_restarting here also to avoid multithreading issues
        self.is_websocket_restarting = True
        # reset might take up to a few seconds, no need to wait for it and block the whole async loop
        threading.Thread(target=self._call_inner_reset, name=f"{self.name}ResetWrapper").start()

    def get_cryptofeed_symbol(self, symbol):
        return symbol

    @classmethod
    def is_supporting_exchange(cls, exchange_candidate_name) -> bool:
        return cls.get_name() == exchange_candidate_name

    def get_pair_from_exchange(self, pair):
        """
        Convert a cryptofeed symbol format to uniformized symbol format
        :param pair: the pair to format
        :return: the formatted pair when success else an empty string
        """
        try:
            return self._get_full_symbol(
                symbol_util.convert_symbol(
                    symbol=pair,
                    symbol_separator=self.CRYPTOFEED_DEFAULT_MARKET_SEPARATOR,
                    new_symbol_separator=octobot_commons.MARKET_SEPARATOR,
                    should_uppercase=True,
                )
            )
        except Exception:
            self.logger.error(f"Failed to get market of {pair}")
        return ""

    def _get_full_symbol(self, feed_symbol):
        try:
            return self._full_symbol_by_feed_symbol[feed_symbol]
        except KeyError:
            return feed_symbol

    def get_exchange_pair(self, pair) -> str:
        """
        Convert an uniformized symbol format to a cryptofeed symbol format
        :param pair: the pair to format
        :return: the cryptofeed pair when success else an empty string
        """
        try:
            feed_symbol = symbol_util.convert_symbol(
                symbol=pair,
                symbol_separator=octobot_commons.MARKET_SEPARATOR,
                new_symbol_separator=self.CRYPTOFEED_DEFAULT_MARKET_SEPARATOR,
                should_uppercase=True,
                base_and_quote_only=True,
            )
            self._full_symbol_by_feed_symbol[feed_symbol] = pair
            return feed_symbol
        except Exception:
            self.logger.error(f"Failed to get market of {pair}")
        return ""

    """
    Private methods
    """

    def _create_client(self):
        """
        Creates cryptofeed client instance
        """
        self._create_client_config()
        self.client = cryptofeed.FeedHandler(config=self.client_config)

    def _create_client_config(self):
        """
        Creates cryptofeed config
        """
        self.client_config = cryptofeed_config.Config()
        try:
            # Disable cryptofeed log file
            self.client_config.config['log']['filename'] = os.devnull
            self.client_config.config['rest']['log']['filename'] = os.devnull
        except KeyError:
            pass
        self.client_config.config[self.get_feed_name().lower()] = self._get_credentials_config()

    def _get_credentials_config(self):
        """
        Add exchange credentials to FeedHandler client config
        """
        if self._should_use_authenticated_feeds():
            key_id, key_secret, key_passphrase = self.get_exchange_credentials()
            return {
                "key_id": key_id,
                "key_secret": key_secret,
                "key_passphrase": key_passphrase
            }
        return {}

    def _init_client(self):
        """
        Prepares client configuration and instantiates client feeds
        """
        try:
            self.client.config = self.client_config
            self._filter_exchange_pairs_and_timeframes()
            self._subscribe_feeds()
        except Exception as e:
            self.logger.exception(e, True, f"Failed to subscribe when creating websocket feed : {e}")

    def _start_client(self, should_create_loop=True):
        """
        Creates client async loop and start client
        :param should_create_loop: When True creates a new async loop. When False uses the current one
        """
        try:
            if should_create_loop:
                # without this two lines `There is no current event loop in thread`
                # is raised when calling `client.run()`
                self.local_loop = asyncio.new_event_loop()
                asyncio.set_event_loop(self.local_loop)
            self.logger.debug(f"Starting websocket client with {len(self.client.feeds)} feeds.")
            self.client.run(start_loop=should_create_loop)
        except Exception as e:
            self.logger.error(f"Failed to start websocket feed : {e}")

    def _subscribe_feeds(self):
        """
        Prepares the subscription of unauthenticated and authenticated feeds and subscribe all
        """
        if self._should_run_candle_feed():
            if self.min_timeframe is None:
                self.logger.error(
                    f"Missing min_timeframe in exchange websocket with candle feeds. This probably means that no "
                    f"required time frame is supported by this exchange's websocket "
                    f"(valid_candle_intervals: {self.cryptofeed_exchange.valid_candle_intervals})")
            self.candle_callback = self.callback_by_feed[self.EXCHANGE_FEEDS[Feeds.CANDLE]]
            self._subscribe_candle_feed()

        # drop unsupported channels
        self.channels = [channel for channel in self.channels
                         if self._is_supported_channel(channel)
                         and channel != self.EXCHANGE_FEEDS.get(Feeds.CANDLE)]

        self.callbacks = {
            channel: self.callback_by_feed[channel]
            for channel in self.channels
            if self.callback_by_feed.get(channel)
        }

        # Add candle callback to callbacks
        try:
            self.callbacks[self.EXCHANGE_FEEDS[Feeds.CANDLE]] = self.candle_callback
        except KeyError:
            pass  # ignore candle callback when Candles feed is not supported

        self._subscribe_channels_feeds()

    def _should_use_authenticated_feeds(self):
        """
        :return: True when authenticated feeds shouldn't be added
        """
        return self._should_authenticate() and self.exchange.authenticated() and not self.exchange_manager.is_sandboxed

    def _is_supported_channel(self, channel):
        """
        Checks if the channel is supported
        :param channel: the channel name
        :return: True if the channel is not unsupported, not ignored
        and if it's an authenticated channel if the exchange is authenticated
        """
        if self.cryptofeed_exchange.is_authenticated_channel(channel) and not self._should_use_authenticated_feeds():
            return False
        return channel != Feeds.UNSUPPORTED.value \
               and not self.should_ignore_feed(self.CRYPTOFEED_FEEDS_TO_WEBSOCKET_FEEDS[channel])

    def _subscribe_candle_feed(self):
        """
        Subscribes a new candle feed for each time frame
        """
        for time_frame in self.time_frames:
            try:
                self._subscribe_feed(
                    symbols=self.filtered_pairs,
                    candle_interval=time_frame.value,
                    channels=self.CANDLE_CHANNELS,
                    callbacks={cryptofeed_constants.CANDLES: self.candle_callback}
                )
            except ValueError as e:
                self.logger.exception(e, True,
                                      f"Error when subscribing to feed: ignored candle feed with {self.filtered_pairs} "
                                      f"on {time_frame.value} ({e})")

    def _subscribe_channels_feeds(self):
        """
        Subscribes all time frame unrelated feeds
        """
        self._subscribe_pair_independent_feed()
        self._subscribe_traded_pairs_feed()
        self._subscribe_watched_pairs_feed()

    def _subscribe_pair_independent_feed(self):
        """
        Subscribes all pair unrelated feeds
        """
        channels = [channel for channel in self.channels if self._is_pair_independent_feed(channel)]
        if channels:
            self._subscribe_feed(
                channels=channels,
                callbacks=self.callbacks
            )

    def _subscribe_traded_pairs_feed(self):
        """
        Subscribes all time frame unrelated feeds for traded pairs
        """
        channels = [channel for channel in self.channels if not self._is_pair_independent_feed(channel)]
        if channels:
            self._subscribe_feed(
                symbols=self.filtered_pairs,
                channels=channels,
                callbacks=self.callbacks
            )

    def _subscribe_watched_pairs_feed(self):
        """
        Subscribes feeds for watched pairs (only on one timeframe for multiple timeframes feeds)
        """
        channels = [channel
                    for channel in self.WATCHED_PAIR_CHANNELS
                    if self._is_supported_channel(channel) and (channel in self.channels or not self.channels)]
        if self.watched_pairs and channels:
            self._subscribe_feed(
                symbols=self.watched_pairs,
                channels=channels,
                callbacks=self.callbacks
            )

    def _get_feed_default_kwargs(self):
        kwargs = {
            "candle_closed_only": False,
            "sandbox": self.exchange_manager.is_sandboxed
        }
        # apply exchange kwargs
        kwargs.update(self.EXCHANGE_CONSTRUCTOR_KWARGS)
        return kwargs

    def _subscribe_feed(self, channels, callbacks, symbols=None, candle_interval=None):
        """
        Subscribe a new feed
        :param symbols: the feed symbols
        :param candle_interval: the feed candle_interval
        :param channels: the feed channels
        :param callbacks: the feed callbacks
        """
        feed_kwargs = self._get_feed_default_kwargs()
        if symbols:
            feed_kwargs["symbols"] = symbols
        if candle_interval:
            # always specify candle_interval even if not always used
            feed_kwargs["candle_interval"] = candle_interval or \
                                             self.min_timeframe.value if self.min_timeframe else candle_interval
        self.client.add_feed(self.get_feed_name(),
                             log_message_on_error=True,
                             callbacks=callbacks,
                             channels=channels,
                             **feed_kwargs)
        for channel in channels:
            symbols_str = f"for {', '.join(symbols)} " if symbols else ""
            candle_interval_str = f"on {candle_interval}" if candle_interval else ""
            self.logger.debug(f"Subscribed to {channel} {symbols_str}{candle_interval_str}")

    def _filter_exchange_pairs_and_timeframes(self):
        """
        Populates self.filtered_pairs and self.min_timeframe
        """
        self._filter_exchange_symbols()
        self._filter_exchange_time_frames()

    def _add_pair(self, pair, watching_only):
        """
        Add a pair to self.filtered_pairs if supported
        :param pair: the pair to add
        :param watching_only: when True add pair to watched_pairs else to filtered_pairs
        """
        if self._is_supported_pair(pair):
            if watching_only:
                self.watched_pairs.append(pair)
            else:
                self.filtered_pairs.append(pair)
        else:
            self.logger.error(f"{self.get_pair_from_exchange(pair)} pair is not supported by this exchange's websocket")

    def _filter_exchange_symbols(self):
        """
        Populates self.filtered_pairs from self.pairs when pair is supported by the cryptofeed exchange
        """
        for pair in self.pairs:
            self._add_pair(pair, watching_only=False)

    def _add_time_frame(self, filtered_timeframes, time_frame, log_on_error):
        """
        Add a time frame to filtered_timeframes if supported
        :param time_frame: the time frame to add
        """
        try:
            if self.cryptofeed_exchange.valid_candle_intervals is not NotImplemented and \
                    time_frame.value in self.cryptofeed_exchange.valid_candle_intervals :
                filtered_timeframes.append(time_frame)
            elif log_on_error:
                self.logger.error(f"{time_frame.value} time frame is not supported by this exchange's websocket")
        except AttributeError:
            # exchange.valid_candle_intervals is not implemented in each cryptofeed exchange
            pass

    def _filter_exchange_time_frames(self):
        """
        Populates self.min_timeframe from self.time_frames when time frame is supported by the cryptofeed exchange
        Leaves self.min_timeframe at None if no timeframe is available on this exchange ws
        """
        # Log error only if necessary (self.min_timeframe is used by candle feed only)
        log_on_error = self._should_run_candle_feed()
        filtered_timeframes = []
        for time_frame in self.time_frames:
            self._add_time_frame(filtered_timeframes, time_frame, log_on_error)
        if filtered_timeframes:
            self.min_timeframe = time_frame_manager.find_min_time_frame(filtered_timeframes)

    def _should_run_candle_feed(self):
        return self.EXCHANGE_FEEDS.get(Feeds.CANDLE) and self.EXCHANGE_FEEDS.get(
            Feeds.CANDLE) != Feeds.UNSUPPORTED.value

    def _is_supported_pair(self, pair):
        return pair in self.cryptofeed_exchange.normalized_symbol_mapping

    def _is_supported_time_frame(self, time_frame):
        return time_frame.value in self.cryptofeed_exchange.valid_candle_intervals

    def _is_pair_independent_feed(self, feed):
        return feed in self.PAIR_INDEPENDENT_CHANNELS

    def _convert_book_prices_to_orders(self, book_prices, book_side):
        """
        Convert a book_prices format : {PRICE_1: SIZE_1, PRICE_2: SIZE_2...}
        to OctoBot's order book format
        :param book_prices: an order book dictionary (order_book.SortedDict)
        :param book_side: a TradeOrderSide value
        :return: the list of order book data converted
        """
        return [
            {
                ECOBIC.PRICE.value: float(order_price),
                ECOBIC.SIZE.value: float(order_size),
                ECOBIC.SIDE.value: book_side,
            }
            for order_price, order_size in book_prices.to_dict().items()
        ]

    def _set_async_callbacks(self):
        """
        Prevent `inspect.iscoroutinefunction` to return False when callback are cythonized
        """
        for callback in self.callback_by_feed.values():
            callback.is_async = True

    def _remove_all_feeds(self):
        """
        Call remove_feed for each client feeds
        """
        for feed in self.client.feeds:
            self._remove_feed(feed)

    def _remove_feed(self, feed):
        """
        Stops and removes a feed from running feeds
        :param feed: the feed instance to remove
        """
        try:
            feed.stop()
            self.client.feeds.remove(feed)
        except ValueError as ve:
            self.logger.error(f"Failed to remove feed from the list of feed : {ve}")
        except Exception as e:
            self.logger.error(f"Failed remove feed : {e}")

    def _fix_signal_handler(self):
        """
        Websocket are started in a new thread thus signal handle cannot be used
        add_signal_handler() can only be called from the main thread
        """
        cryptofeed.feedhandler.SIGNALS = ()

    def _fix_logger(self):
        """
        Replace cryptofeed feedhandler logger because it writes logs into "cryptofeed.log" file
        """
        cryptofeed.feedhandler.LOG = self.client_logger
        cryptofeed.connection_handler.LOG = self.client_logger

    """
    Callbacks
    """

    async def ticker(self, ticker: cryptofeed_types.Ticker, receipt_timestamp: float):
        """
        Cryptofeed ticker callback
        :param ticker: the ticker object defined in cryptofeed.types.Ticker
        :param receipt_timestamp: received timestamp
        """
        symbol = self.get_pair_from_exchange(ticker.symbol)
        # Can't create a full ticker from bid, ask, timestamp and symbol data
        # Push (ask + bid) / 2 as close price in MARK_PRICE channel
        await self.push_to_channel(trading_constants.MARK_PRICE_CHANNEL,
                                   symbol,
                                   (ticker.ask + ticker.bid) / decimal.Decimal(2),
                                   mark_price_source=trading_enums.MarkPriceSources.TICKER_CLOSE_PRICE.value)

    async def trades(self, trade: cryptofeed_types.Trade, receipt_timestamp: float):
        """
        Cryptofeed ticker callback
        :param trade: the trade object defined in cryptofeed.types.Trade
        :param receipt_timestamp: received timestamp
        """
        symbol = self.get_pair_from_exchange(trade.symbol)
        await self.push_to_channel(trading_constants.RECENT_TRADES_CHANNEL,
                                   symbol,
                                   [{
                                       trading_enums.ExchangeConstantsOrderColumns.TIMESTAMP.value: trade.timestamp,
                                       trading_enums.ExchangeConstantsOrderColumns.SYMBOL.value: symbol,
                                       trading_enums.ExchangeConstantsOrderColumns.ID.value: trade.id,
                                       trading_enums.ExchangeConstantsOrderColumns.TYPE.value: None,
                                       trading_enums.ExchangeConstantsOrderColumns.SIDE.value: trade.side,
                                       trading_enums.ExchangeConstantsOrderColumns.PRICE.value: float(trade.price),
                                       trading_enums.ExchangeConstantsOrderColumns.AMOUNT.value: float(trade.amount)
                                   }])

    async def book(self, order_book: cryptofeed_types.OrderBook, receipt_timestamp: float):
        """
        Cryptofeed orderbook callback
        :param order_book: the order_book object defined in cryptofeed.types.OrderBook
        :param receipt_timestamp: received timestamp
        """
        symbol = self.get_pair_from_exchange(order_book.symbol)
        book_instance = self.get_book_instance(symbol)

        book_instance.handle_book_adds(
            self._convert_book_prices_to_orders(
                book_prices=order_book.book.asks,
                book_side=trading_enums.TradeOrderSide.SELL.value) +
            self._convert_book_prices_to_orders(
                book_prices=order_book.book.bids,
                book_side=trading_enums.TradeOrderSide.BUY.value))

        await self.push_to_channel(trading_constants.ORDER_BOOK_CHANNEL,
                                   symbol,
                                   book_instance.asks,
                                   book_instance.bids,
                                   update_order_book=False)

    async def candle(self, candle_data: cryptofeed_types.Candle, receipt_timestamp: float):
        """
        Cryptofeed candle callback
        :param candle_data: the candle object defined in cryptofeed.types.Candle
        :param receipt_timestamp: received timestamp
        """
        symbol = self.get_pair_from_exchange(candle_data.symbol)
        time_frame = commons_enums.TimeFrames(candle_data.interval)
        candle = [
            candle_data.start,
            float(candle_data.open),
            float(candle_data.high),
            float(candle_data.low),
            float(candle_data.close),
            float(candle_data.volume),
        ]
        ticker = {
            Ectc.HIGH.value: float(candle_data.high),
            Ectc.LOW.value: float(candle_data.low),
            Ectc.BID.value: None,
            Ectc.BID_VOLUME.value: None,
            Ectc.ASK.value: None,
            Ectc.ASK_VOLUME.value: None,
            Ectc.OPEN.value: float(candle_data.open),
            Ectc.CLOSE.value: float(candle_data.close),
            Ectc.LAST.value: float(candle_data.close),
            Ectc.PREVIOUS_CLOSE.value: None,
            Ectc.BASE_VOLUME.value: float(candle_data.volume),
            Ectc.TIMESTAMP.value: self.exchange.get_exchange_current_time(),
        }

        if candle_data.symbol not in self.watched_pairs:
            previous_candle = self._get_previous_open_candle(time_frame, symbol)
            push_previous_candle = previous_candle is not None and \
                previous_candle[commons_enums.PriceIndexes.IND_PRICE_TIME.value] < candle_data.start
            if candle_data.closed or push_previous_candle:
                if candle_data.closed and previous_candle is not None and \
                   previous_candle[commons_enums.PriceIndexes.IND_PRICE_TIME.value] > candle_data.start:
                    self.logger.warning(f"Duplicate closed candle: pushing already pushed closed "
                                        f"candle: [{candle_data}]")
                await self.push_to_channel(trading_constants.OHLCV_CHANNEL,
                                           time_frame,
                                           symbol,
                                           previous_candle if push_previous_candle else candle)
                # closed candle has been fetched from exchange, use it and reset previous open candle
                self._register_previous_open_candle(time_frame, symbol, None)
            if not candle_data.closed:
                await self.push_to_channel(trading_constants.KLINE_CHANNEL,
                                           time_frame,
                                           symbol,
                                           candle)
                self._register_previous_open_candle(time_frame, symbol, candle)

        # Push a new ticker if necessary : only push on the min timeframe
        if time_frame is self.min_timeframe:
            await self.push_to_channel(trading_constants.TICKER_CHANNEL,
                                       symbol,
                                       ticker)

    async def liquidations(self, liquidation: cryptofeed_types.Liquidation, receipt_timestamp: float):
        """
        Cryptofeed liquidation callback
        :param liquidation: the liquidation object defined in cryptofeed.types.Liquidation
        :param receipt_timestamp: received timestamp
        """

    async def funding(self, funding: cryptofeed_types.Funding, receipt_timestamp: float):
        """
        Cryptofeed funding callback
        :param funding: the funding object defined in cryptofeed.types.Funding
        :param receipt_timestamp: received timestamp
        """
        symbol = self.get_pair_from_exchange(funding.symbol)
        await self.push_to_channel(trading_constants.FUNDING_CHANNEL,
                                   symbol,
                                   funding.rate,
                                   funding.predicted_rate,
                                   funding.next_funding_time,
                                   funding.timestamp)
        await self.push_to_channel(trading_constants.MARK_PRICE_CHANNEL,
                                   symbol,
                                   funding.mark_price)

    async def open_interest(self, open_interest: cryptofeed_types.OpenInterest, receipt_timestamp: float):
        """
        Cryptofeed open interest callback
        :param open_interest: the open_interest object defined in cryptofeed.types.OpenInterest
        :param receipt_timestamp: received timestamp
        """

    async def index(self, index: cryptofeed_types.Index, receipt_timestamp: float):
        """
        Cryptofeed future index callback
        :param index: the index object defined in cryptofeed.types.Index
        :param receipt_timestamp: received timestamp
        """
        await self.push_to_channel(trading_constants.MARK_PRICE_CHANNEL,
                                   self.get_pair_from_exchange(index.symbol),
                                   index.price,
                                   trading_enums.MarkPriceSources.EXCHANGE_MARK_PRICE.value)

    async def order(self, order: cryptofeed_types.OrderInfo, receipt_timestamp: float):
        """
        Cryptofeed order callback
        :param order the order object defined in cryptofeed.types.?
        :param receipt_timestamp: received timestamp
        """
        await self.push_to_channel(trading_constants.ORDERS_CHANNEL, [{
            trading_enums.ExchangeConstantsOrderColumns.TYPE.value: self._parse_order_type(order.type),
            trading_enums.ExchangeConstantsOrderColumns.STATUS.value: self._parse_order_status(order.status),
            trading_enums.ExchangeConstantsOrderColumns.ID.value: order.id,
            trading_enums.ExchangeConstantsOrderColumns.SYMBOL.value: order.symbol,
            trading_enums.ExchangeConstantsOrderColumns.PRICE.value: order.price,
            trading_enums.ExchangeConstantsOrderColumns.AMOUNT.value: order.amount,
            trading_enums.ExchangeConstantsOrderColumns.REMAINING.value: order.remaining,
            trading_enums.ExchangeConstantsOrderColumns.FILLED.value: order.amount - order.remaining,
            trading_enums.ExchangeConstantsOrderColumns.TIMESTAMP.value: order.timestamp,
            trading_enums.ExchangeConstantsOrderColumns.SIDE.value: self._parse_order_side(order.side)
        }])

    async def trade(self, trade, receipt_timestamp: float):
        """
        Cryptofeed filled order callback
        :param trade: the trade object defined in cryptofeed.types.?
        :param receipt_timestamp: received timestamp
        """
        await self.push_to_channel(trading_constants.TRADES_CHANNEL, [self.exchange.parse_trade(trade)])

    async def balance(self, balance: cryptofeed_types.Balance, receipt_timestamp: float):
        """
        Cryptofeed balance callback
        :param balance: the balance object defined in cryptofeed.types.Balance
        :param receipt_timestamp: received timestamp
        """
        await self.push_to_channel(trading_constants.BALANCE_CHANNEL,
                                   self.exchange.parse_balance(balance.balance))

    async def transaction(self, transaction: cryptofeed_types.Transaction, receipt_timestamp: float):
        """
        Cryptofeed transaction callback
        :param transaction: the transaction object defined in cryptofeed.types.Transaction
        :param receipt_timestamp: received timestamp
        """

    async def fill(self, fill: cryptofeed_types.Fill, receipt_timestamp: float):
        """
        Cryptofeed fill callback
        :param fill: the fill object defined in cryptofeed.types.Fill
        :param receipt_timestamp: received timestamp
        """

    """
    Parsers
    """

    def _parse_order_type(self, raw_order_type):
        """
        :param raw_order_type: the cryptofeed unified order type
        :return: the unified OctoBot order type
        """
        # TODO Migrate to switch case with python 3.10
        if raw_order_type == cryptofeed_constants.LIMIT:
            return trading_enums.TradeOrderType.LIMIT.value
        elif raw_order_type == cryptofeed_constants.MARKET:
            return trading_enums.TradeOrderType.MARKET.value
        elif raw_order_type == cryptofeed_constants.STOP_LIMIT:
            return trading_enums.TradeOrderType.STOP_LOSS_LIMIT.value
        elif raw_order_type == cryptofeed_constants.STOP_MARKET:
            return trading_enums.TradeOrderType.STOP_LOSS.value
        elif raw_order_type == cryptofeed_constants.MAKER_OR_CANCEL:
            pass
        elif raw_order_type == cryptofeed_constants.FILL_OR_KILL:
            pass
        elif raw_order_type == cryptofeed_constants.IMMEDIATE_OR_CANCEL:
            pass
        elif raw_order_type == cryptofeed_constants.GOOD_TIL_CANCELED:
            pass
        self.logger.error(f"Failed to parse {raw_order_type} order type (not supported).")
        return None

    def _parse_order_status(self, raw_order_status):
        """
        :param raw_order_status: the cryptofeed unified order status
        :return: the unified OctoBot order status
        """
        # TODO Migrate to switch case with python 3.10
        if raw_order_status == cryptofeed_constants.OPEN:
            return trading_enums.OrderStatus.OPEN.value
        elif raw_order_status == cryptofeed_constants.PENDING:
            return trading_enums.OrderStatus.OPEN.value
        elif raw_order_status == cryptofeed_constants.FILLED:
            return trading_enums.OrderStatus.FILLED.value
        elif raw_order_status == cryptofeed_constants.PARTIAL:
            return trading_enums.OrderStatus.PARTIALLY_FILLED.value
        elif raw_order_status == cryptofeed_constants.CANCELLED:
            return trading_enums.OrderStatus.CANCELED.value
        elif raw_order_status == cryptofeed_constants.UNFILLED:
            return trading_enums.OrderStatus.OPEN.value
        elif raw_order_status == cryptofeed_constants.EXPIRED:
            return trading_enums.OrderStatus.EXPIRED.value
        elif raw_order_status == cryptofeed_constants.FAILED:
            return trading_enums.OrderStatus.REJECTED.value
        elif raw_order_status == cryptofeed_constants.SUBMITTING:
            return trading_enums.OrderStatus.OPEN.value
        elif raw_order_status == cryptofeed_constants.CANCELLING:
            return trading_enums.OrderStatus.PENDING_CANCEL.value
        elif raw_order_status == cryptofeed_constants.CLOSED:
            return trading_enums.OrderStatus.CLOSED.value
        elif raw_order_status == cryptofeed_constants.SUSPENDED:
            pass
        self.logger.error(f"Failed to parse {raw_order_status} order status (not supported).")
        return None

    def _parse_order_side(self, raw_order_side):
        """
        :param raw_order_side: the cryptofeed unified order side
        :return: the unified OctoBot order side
        """
        return trading_enums.TradeOrderSide.BUY.value \
            if raw_order_side == cryptofeed_constants.BUY else trading_enums.TradeOrderSide.SELL.value

    def _register_previous_open_candle(self, time_frame, symbol, candle):
        try:
            self._previous_open_candles[time_frame][symbol] = candle
        except KeyError:
            if time_frame not in self._previous_open_candles:
                self._previous_open_candles[time_frame] = {}
            self._previous_open_candles[time_frame][symbol] = candle

    def _get_previous_open_candle(self, time_frame, symbol):
        try:
            return self._previous_open_candles[time_frame][symbol]
        except KeyError:
            return None



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data cimport state
from octobot_trading.personal_data.state cimport (
    State,
)
from octobot_trading.personal_data cimport orders
from octobot_trading.personal_data.orders cimport (
    Order,
    OrderState,
    OrdersManager,
    UnknownOrder,
    MarketOrder,
    SellMarketOrder,
    BuyMarketOrder,
    BuyLimitOrder,
    SellLimitOrder,
    LimitOrder,
    TakeProfitOrder,
    StopLossOrder,
    StopLossLimitOrder,
    TakeProfitLimitOrder,
    TrailingStopOrder,
    TrailingStopLimitOrder,
)
from octobot_trading.personal_data cimport portfolios
from octobot_trading.personal_data.portfolios cimport (
    BalanceUpdaterSimulator,
    BalanceProfitabilityUpdaterSimulator,
    create_portfolio_from_exchange_manager,
    BalanceUpdater,
    BalanceProfitabilityUpdater,
    Portfolio,
    BalanceProducer,
    BalanceChannel,
    BalanceProfitabilityProducer,
    BalanceProfitabilityChannel,
    SubPortfolio,
    PortfolioManager,
    FuturePortfolio,
    MarginPortfolio,
    SpotPortfolio,
    Asset,
    FutureAsset,
    MarginAsset,
    SpotAsset,
    parse_decimal_portfolio,
    parse_decimal_config_portfolio,
    portfolio_to_float,
)
from octobot_trading.personal_data cimport positions
from octobot_trading.personal_data.positions cimport (
    PositionState,
    PositionsProducer,
    PositionsChannel,
    PositionsUpdaterSimulator,
    Position,
    parse_position_type,
    CrossPosition,
    IsolatedPosition,
    PositionsUpdater,
    PositionsManager,
    create_position_instance_from_raw,
    create_position_from_type,
    create_symbol_position,
    parse_position_status,
    LiquidatePositionState,
    OpenPositionState,
)
from octobot_trading.personal_data cimport trades
from octobot_trading.personal_data.trades cimport (
    TradesManager,
    TradesProducer,
    TradesChannel,
    create_trade_instance_from_raw,
    create_trade_from_order,
    create_trade_instance,
    TradesUpdater,
    Trade,
)
from octobot_trading.personal_data cimport transactions
from octobot_trading.personal_data.transactions cimport (
    TransactionsManager,
    Transaction,
    BlockchainTransaction,
    FeeTransaction,
    RealisedPnlTransaction,
    TransferTransaction,
    create_blockchain_transaction,
    create_realised_pnl_transaction,
    create_fee_transaction,
    create_transfer_transaction,
)
from octobot_trading.personal_data cimport exchange_personal_data
from octobot_trading.personal_data.exchange_personal_data cimport (
    ExchangePersonalData,
)

__all__ = [
    "State",
    "Order",
    "OrderState",
    "OrdersUpdater",
    "OrdersProducer",
    "OrdersChannel",
    "OrdersManager",
    "OrdersUpdaterSimulator",
    "CloseOrderState",
    "CancelOrderState",
    "OpenOrderState",
    "FillOrderState",
    "UnknownOrder",
    "MarketOrder",
    "SellMarketOrder",
    "BuyMarketOrder",
    "BuyLimitOrder",
    "SellLimitOrder",
    "LimitOrder",
    "TakeProfitOrder",
    "StopLossOrder",
    "StopLossLimitOrder",
    "TakeProfitLimitOrder",
    "TrailingStopOrder",
    "TrailingStopLimitOrder",
    "PositionsProducer",
    "PositionsChannel",
    "BalanceUpdaterSimulator",
    "BalanceProfitabilityUpdaterSimulator",
    "create_portfolio_from_exchange_manager",
    "BalanceUpdater",
    "BalanceProfitabilityUpdater",
    "Portfolio",
    "BalanceProducer",
    "BalanceChannel",
    "BalanceProfitabilityProducer",
    "BalanceProfitabilityChannel",
    "SubPortfolio",
    "PortfolioManager",
    "FuturePortfolio",
    "MarginPortfolio",
    "SpotPortfolio",
    "Asset",
    "FutureAsset",
    "MarginAsset",
    "SpotAsset",
    "parse_decimal_portfolio",
    "parse_decimal_config_portfolio",
    "portfolio_to_float",
    "PositionsUpdaterSimulator",
    "Position",
    "parse_position_type",
    "CrossPosition",
    "IsolatedPosition",
    "PositionState",
    "LiquidatePositionState",
    "OpenPositionState",
    "PositionsUpdater",
    "PositionsManager",
    "create_position_instance_from_raw",
    "create_position_from_type",
    "create_symbol_position",
    "parse_position_status",
    "TradesManager",
    "TradesProducer",
    "TradesChannel",
    "create_trade_instance_from_raw",
    "create_trade_from_order",
    "create_trade_instance",
    "TradesUpdater",
    "Trade",
    "ExchangePersonalData",
    "TransactionsManager",
    "Transaction",
    "BlockchainTransaction",
    "FeeTransaction",
    "RealisedPnlTransaction",
    "TransferTransaction",
    "create_blockchain_transaction",
    "create_realised_pnl_transaction",
    "create_fee_transaction",
    "create_transfer_transaction",
]



# pylint: disable=E0203
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_commons.tree as commons_tree

import async_channel.enums as channel_enums
import async_channel.constants as channel_constants
import async_channel.channels as channels
import async_channel.consumer as consumers
import async_channel.producer as producers

import octobot_commons.logging as logging


class ExchangeChannelConsumer(consumers.Consumer):
    """
    Consumer adapted for ExchangeChannel
    """


class ExchangeChannelInternalConsumer(consumers.InternalConsumer):
    """
    InternalConsumer adapted for ExchangeChannel
    """


class ExchangeChannelSupervisedConsumer(consumers.SupervisedConsumer):
    """
    SupervisedConsumer adapted for ExchangeChannel
    """


class ExchangeChannelProducer(producers.Producer):
    """
    Producer adapted for ExchangeChannel
    """

    def __init__(self, channel):
        super().__init__(channel)
        self.logger = logging.get_logger(f"{self.__class__.__name__}[{channel.exchange_manager.exchange_name}]")

    async def fetch_and_push(self):
        self.logger.error("self.fetch_and_push() is not implemented")

    def trigger_single_update(self):
        asyncio.create_task(self.fetch_and_push())

    async def wait_for_dependencies(self, paths, timeout):
        for path in paths:
            if not await commons_tree.EventProvider.instance().wait_for_event(
                self.channel.exchange_manager.bot_id,
                path,
                timeout
            ):
                return False
        return True


class ExchangeChannel(channels.Channel):
    PRODUCER_CLASS = ExchangeChannelProducer
    CONSUMER_CLASS = ExchangeChannelConsumer
    CRYPTOCURRENCY_KEY = "cryptocurrency"
    SYMBOL_KEY = "symbol"
    DEFAULT_PRIORITY_LEVEL = channel_enums.ChannelConsumerPriorityLevels.HIGH.value

    def __init__(self, exchange_manager):
        super().__init__()
        self.logger = logging.get_logger(f"{self.__class__.__name__}[{exchange_manager.exchange_name}]")
        self.exchange_manager = exchange_manager
        self.exchange = exchange_manager.exchange

        self.filter_send_counter = 0
        self.should_send_filter = False

    async def new_consumer(self,
                           callback: object = None,
                           consumer_instance: object = None,
                           size: int = 0,
                           priority_level: int = DEFAULT_PRIORITY_LEVEL,
                           symbol: str = channel_constants.CHANNEL_WILDCARD,
                           cryptocurrency: str = channel_constants.CHANNEL_WILDCARD,
                           **kwargs) -> ExchangeChannelConsumer:
        consumer = consumer_instance if consumer_instance else self.CONSUMER_CLASS(callback,
                                                                                   size=size,
                                                                                   priority_level=priority_level)
        await self._add_new_consumer_and_run(consumer,
                                             cryptocurrency=cryptocurrency,
                                             symbol=symbol,
                                             **kwargs)
        await self._check_producers_state()
        return consumer

    def get_filtered_consumers(self,
                               cryptocurrency=channel_constants.CHANNEL_WILDCARD,
                               symbol=channel_constants.CHANNEL_WILDCARD):
        return self.get_consumer_from_filters({
            self.CRYPTOCURRENCY_KEY: cryptocurrency,
            self.SYMBOL_KEY: symbol
        })

    async def _add_new_consumer_and_run(self, consumer,
                                        cryptocurrency=channel_constants.CHANNEL_WILDCARD,
                                        symbol=channel_constants.CHANNEL_WILDCARD):
        self.add_new_consumer(consumer,
                              {
                                  self.CRYPTOCURRENCY_KEY: cryptocurrency,
                                  self.SYMBOL_KEY: symbol
                              })
        await self._run_consumer(consumer,
                                 symbol=symbol)

    async def _run_consumer(self, consumer,
                            symbol=channel_constants.CHANNEL_WILDCARD):
        await consumer.run(with_task=not self.is_synchronized)
        self.logger.debug(f"Consumer started for symbol {symbol}: {consumer}")


class TimeFrameExchangeChannel(ExchangeChannel):
    TIME_FRAME_KEY = "time_frame"

    def get_filtered_consumers(self,
                               cryptocurrency=channel_constants.CHANNEL_WILDCARD,
                               symbol=channel_constants.CHANNEL_WILDCARD,
                               time_frame=channel_constants.CHANNEL_WILDCARD):
        return self.get_consumer_from_filters({
            self.CRYPTOCURRENCY_KEY: cryptocurrency,
            self.SYMBOL_KEY: symbol,
            self.TIME_FRAME_KEY: time_frame
        })

    async def _add_new_consumer_and_run(self, consumer,
                                        cryptocurrency=channel_constants.CHANNEL_WILDCARD,
                                        symbol=channel_constants.CHANNEL_WILDCARD,
                                        time_frame=channel_constants.CHANNEL_WILDCARD):
        self.add_new_consumer(consumer,
                              {
                                  self.CRYPTOCURRENCY_KEY: cryptocurrency,
                                  self.SYMBOL_KEY: symbol,
                                  self.TIME_FRAME_KEY: time_frame
                              })
        await self._run_consumer(consumer,
                                 symbol=symbol)


def set_chan(chan, name) -> None:
    chan_name = chan.get_name() if name else name

    try:
        exchange_chan = channels.ChannelInstances.instance().channels[chan.exchange_manager.id]
    except KeyError:
        channels.ChannelInstances.instance().channels[chan.exchange_manager.id] = {}
        exchange_chan = channels.ChannelInstances.instance().channels[chan.exchange_manager.id]

    if chan_name not in exchange_chan:
        exchange_chan[chan_name] = chan
    else:
        raise ValueError(f"Channel {chan_name} already exists.")


def get_exchange_channels(exchange_id) -> dict:
    try:
        return channels.ChannelInstances.instance().channels[exchange_id]
    except KeyError:
        raise KeyError(f"Channels not found on exchange with id: {exchange_id}")


def del_exchange_channel_container(exchange_id):
    try:
        channels.ChannelInstances.instance().channels.pop(exchange_id, None)
    except KeyError:
        raise KeyError(f"Channels not found on exchange with id: {exchange_id}")


def get_chan(chan_name, exchange_id) -> ExchangeChannel:
    try:
        return channels.ChannelInstances.instance().channels[exchange_id][chan_name]
    except KeyError:
        # get_logger(ExchangeChannel.__name__).error(f"Channel {chan_name} not found on exchange with id: "
        #                                            f"{exchange_id}")
        raise KeyError(f"Channel {chan_name} not found on exchange with id: {exchange_id}")


def del_chan(chan_name, exchange_id) -> None:
    try:
        channels.ChannelInstances.instance().channels[exchange_id].pop(chan_name, None)
    except KeyError:
        logging.get_logger(ExchangeChannel.__name__).warning(f"Can't del chan {chan_name} "
                                                             f"on exchange with id: {exchange_id}")


async def stop_exchange_channels(exchange_manager, should_warn=True) -> None:
    """
    Stop exchange channels and producers
    :param exchange_manager: the related exchange manager
    :param should_warn: if an error message should be logged if an error happened during stopping process
    """
    try:
        for channel_name in list(get_exchange_channels(exchange_manager.id)):
            channel = get_chan(channel_name, exchange_manager.id)
            await channel.stop()
            for consumer in channel.consumers:
                await channel.remove_consumer(consumer)
            get_chan(channel_name, exchange_manager.id).flush()
            del_chan(channel_name, exchange_manager.id)
        del_exchange_channel_container(exchange_manager.id)
    except KeyError:
        if should_warn:
            exchange_manager.logger.error(f"No exchange channel for this exchange (id: {exchange_manager.id})")
